from __future__ import with_statement
import unittest
import inspect
from datetime import datetime

from whoosh import analysis, fields, formats, qparser, query
from whoosh.filedb.filestore import RamStorage
from whoosh.qparser import dateparse
from whoosh.support.times import adatetime


class TestParserPlugins(unittest.TestCase):
    def _plugin_classes(self, ignore):
        # Get all the subclasses of Weighting in whoosh.scoring
        return [c for name, c in inspect.getmembers(qparser, inspect.isclass)
                if qparser.Plugin in c.__bases__ and c not in ignore]
    
    def test_combos(self):
        qs = 'w:a "hi there"^4.2 AND x:b^2.3 OR c AND (y:d OR e) (apple ANDNOT bear)^2.3'
        
        init_args = {qparser.DisMaxPlugin: ({"content": 1.0, "title": 1.2}, ),
                     qparser.FieldAliasPlugin: ({"content": ("text", "body")}, ),
                     qparser.MultifieldPlugin: (["title", "content"], ),
                     qparser.CopyFieldPlugin: ({"name": "phone"}, ),
                     }
        
        plugins = self._plugin_classes(())
        for i, plugin in enumerate(plugins):
            try:
                plugins[i] = plugin(*init_args.get(plugin, ()))
            except TypeError:
                raise TypeError("Error instantiating %s" % plugin)
        
        count = 0
        for i, first in enumerate(plugins):
            for j in xrange(len(plugins)):
                if i == j: continue
                plist = [p for p in plugins[:j] if p is not first] + [first]
                qp = qparser.QueryParser("text", None, plugins=plist)
                try:
                    qp.parse(qs)
                except Exception, e:
                    raise Exception(str(e) + " combo: %s %r" % (count, plist))
                count += 1

    def test_field_alias(self):
        qp = qparser.QueryParser("content", None)
        qp.add_plugin(qparser.FieldAliasPlugin({"title": ("article", "caption")}))
        q = qp.parse("alfa title:bravo article:charlie caption:delta")
        self.assertEqual(unicode(q), u"(content:alfa AND title:bravo AND title:charlie AND title:delta)")

    def test_dateparser(self):
        schema = fields.Schema(text=fields.TEXT, date=fields.DATETIME)
        qp = qparser.QueryParser("text", schema)
        
        errs = []
        def cb(arg):
            errs.append(arg)
        basedate = datetime(2010, 9, 20, 15, 16, 6, 454000)
        qp.add_plugin(dateparse.DateParserPlugin(basedate, callback=cb))
        
        q = qp.parse(u"hello date:'last tuesday'")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(q[1].__class__, query.DateRange)
        self.assertEqual(q[1].startdate, adatetime(2010, 9, 14).floor())
        self.assertEqual(q[1].enddate, adatetime(2010, 9, 14).ceil())
        
        q = qp.parse(u"date:'3am to 5pm'")
        self.assertEqual(q.__class__, query.DateRange)
        self.assertEqual(q.startdate, adatetime(2010, 9, 20, 3).floor())
        self.assertEqual(q.enddate, adatetime(2010, 9, 20, 17).ceil())
        
        q = qp.parse(u"date:blah")
        self.assertEqual(q, query.NullQuery)
        self.assertEqual(errs[0], "blah")
        
        q = qp.parse(u"hello date:blarg")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.fieldname, "text")
        self.assertEqual(q.text, "hello")
        self.assertEqual(errs[1], "blarg")
        
        q = qp.parse(u"hello date:20055x10")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.fieldname, "text")
        self.assertEqual(q.text, "hello")
        self.assertEqual(errs[2], "20055x10")
        
        q = qp.parse(u"hello date:'2005 19 32'")
        self.assertEqual(q.__class__, query.Term)
        self.assertEqual(q.fieldname, "text")
        self.assertEqual(q.text, "hello")
        self.assertEqual(errs[3], "2005 19 32")
        
        q = qp.parse(u"date:'march 24 to dec 12'")
        self.assertEqual(q.__class__, query.DateRange)
        self.assertEqual(q.startdate, adatetime(2010, 3, 24).floor())
        self.assertEqual(q.enddate, adatetime(2010, 12, 12).ceil())
        
        q = qp.parse(u"date:('30 june' OR '10 july') quick")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(len(q), 2)
        self.assertEqual(q[0].__class__, query.Or)
        self.assertEqual(q[0][0].__class__, query.DateRange)
        self.assertEqual(q[0][1].__class__, query.DateRange)
        
    def test_date_range(self):
        schema = fields.Schema(text=fields.TEXT, date=fields.DATETIME)
        qp = qparser.QueryParser("text", schema)
        basedate = datetime(2010, 9, 20, 15, 16, 6, 454000)
        qp.add_plugin(dateparse.DateParserPlugin(basedate))
        
        q = qp.parse(u"date:['30 march' to 'next wednesday']")
        self.assertEqual(q.__class__, query.DateRange)
        self.assertEqual(q.startdate, adatetime(2010, 3, 30).floor())
        self.assertEqual(q.enddate, adatetime(2010, 9, 22).ceil())
        
        q = qp.parse(u"date:[to 'next wednesday']")
        self.assertEqual(q.__class__, query.DateRange)
        self.assertEqual(q.startdate, None)
        self.assertEqual(q.enddate, adatetime(2010, 9, 22).ceil())
        
        q = qp.parse(u"date:['30 march' to]")
        self.assertEqual(q.__class__, query.DateRange)
        self.assertEqual(q.startdate, adatetime(2010, 3, 30).floor())
        self.assertEqual(q.enddate, None)
        
        q = qp.parse(u"date:[30 march to next wednesday]")
        self.assertEqual(q.__class__, query.DateRange)
        self.assertEqual(q.startdate, adatetime(2010, 3, 30).floor())
        self.assertEqual(q.enddate, adatetime(2010, 9, 22).ceil())
        
        q = qp.parse(u"date:[to next wednesday]")
        self.assertEqual(q.__class__, query.DateRange)
        self.assertEqual(q.startdate, None)
        self.assertEqual(q.enddate, adatetime(2010, 9, 22).ceil())
        
        q = qp.parse(u"date:[30 march to]")
        self.assertEqual(q.__class__, query.DateRange)
        self.assertEqual(q.startdate, adatetime(2010, 3, 30).floor())
        self.assertEqual(q.enddate, None)
    
    def test_daterange_empty_field(self):
        schema = fields.Schema(test=fields.DATETIME)
        ix = RamStorage().create_index(schema)
                        
        writer = ix.writer()
        writer.add_document(test=None)
        writer.commit()
        
        with ix.searcher() as s:
            q = query.DateRange("test", datetime.fromtimestamp(0), datetime.today())
            r = s.search(q)
            self.assertEqual(len(r), 0)
    
    def test_free_dates(self):
        a = analysis.StandardAnalyzer(stoplist=None)
        schema = fields.Schema(text=fields.TEXT(analyzer=a), date=fields.DATETIME)
        qp = qparser.QueryParser("text", schema)
        basedate = datetime(2010, 9, 20, 15, 16, 6, 454000)
        qp.add_plugin(dateparse.DateParserPlugin(basedate, free=True))
        
        q = qp.parse(u"hello date:last tuesday")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(len(q), 2)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[0].text, "hello")
        self.assertEqual(q[1].__class__, query.DateRange)
        self.assertEqual(q[1].startdate, adatetime(2010, 9, 14).floor())
        self.assertEqual(q[1].enddate, adatetime(2010, 9, 14).ceil())
        
        q = qp.parse(u"date:mar 29 1972 hello")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(len(q), 2)
        self.assertEqual(q[0].__class__, query.DateRange)
        self.assertEqual(q[0].startdate, adatetime(1972, 3, 29).floor())
        self.assertEqual(q[0].enddate, adatetime(1972, 3, 29).ceil())
        self.assertEqual(q[1].__class__, query.Term)
        self.assertEqual(q[1].text, "hello")

        q = qp.parse(u"date:2005 march 2")
        self.assertEqual(q.__class__, query.DateRange)
        self.assertEqual(q.startdate, adatetime(2005, 3, 2).floor())
        self.assertEqual(q.enddate, adatetime(2005, 3, 2).ceil())
        
        q = qp.parse(u"date:'2005' march 2")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(len(q), 3)
        self.assertEqual(q[0].__class__, query.DateRange)
        self.assertEqual(q[0].startdate, adatetime(2005).floor())
        self.assertEqual(q[0].enddate, adatetime(2005).ceil())
        self.assertEqual(q[1].__class__, query.Term)
        self.assertEqual(q[1].fieldname, "text")
        self.assertEqual(q[1].text, "march")
        
        q = qp.parse(u"date:march 24 to dec 12")
        self.assertEqual(q.__class__, query.DateRange)
        self.assertEqual(q.startdate, adatetime(2010, 3, 24).floor())
        self.assertEqual(q.enddate, adatetime(2010, 12, 12).ceil())
        
        q = qp.parse(u"date:5:10pm")
        self.assertEqual(q.__class__, query.DateRange)
        self.assertEqual(q.startdate, adatetime(2010, 9, 20, 17, 10).floor())
        self.assertEqual(q.enddate, adatetime(2010, 9, 20, 17, 10).ceil())
        
        q = qp.parse(u"(date:30 june OR date:10 july) quick")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(len(q), 2)
        self.assertEqual(q[0].__class__, query.Or)
        self.assertEqual(q[0][0].__class__, query.DateRange)
        self.assertEqual(q[0][1].__class__, query.DateRange)
        
    def test_prefix_plugin(self):
        schema = fields.Schema(id=fields.ID, text=fields.TEXT)
        ix = RamStorage().create_index(schema)
        
        w = ix.writer()
        w.add_document(id=u"1", text=u"alfa")
        w.add_document(id=u"2", text=u"bravo")
        w.add_document(id=u"3", text=u"buono")
        w.commit()
        
        with ix.searcher() as s:
            qp = qparser.QueryParser("text", schema)
            qp.remove_plugin_class(qparser.WildcardPlugin)
            qp.add_plugin(qparser.PrefixPlugin)
            
            q = qp.parse(u"b*")
            r = s.search(q, limit=None)
            self.assertEqual(len(r), 2)
            
            q = qp.parse(u"br*")
            r = s.search(q, limit=None)
            self.assertEqual(len(r), 1)
        
    def test_custom_tokens(self):
        qp = qparser.QueryParser("text", None)
        qp.remove_plugin_class(qparser.CompoundsPlugin)
        qp.remove_plugin_class(qparser.NotPlugin)
        
        cp = qparser.CompoundsPlugin(And="&", Or="\\|", AndNot="&!", AndMaybe="&~", Not=None)
        qp.add_plugin(cp)
        
        np = qparser.NotPlugin("-")
        qp.add_plugin(np)
        
        q = qp.parse("this | that")
        self.assertEqual(q.__class__, query.Or)
        self.assertEqual(q[0].__class__, query.Term)
        self.assertEqual(q[0].text, "this")
        self.assertEqual(q[1].__class__, query.Term)
        self.assertEqual(q[1].text, "that")
        
        q = qp.parse("this&!that")
        self.assertEqual(q.__class__, query.AndNot)
        self.assertEqual(q.a.__class__, query.Term)
        self.assertEqual(q.a.text, "this")
        self.assertEqual(q.b.__class__, query.Term)
        self.assertEqual(q.b.text, "that")
        
        q = qp.parse("alfa -bravo NOT charlie")
        self.assertEqual(len(q), 4)
        self.assertEqual(q[1].__class__, query.Not)
        self.assertEqual(q[1].query.text, "bravo")
        self.assertEqual(q[2].text, "NOT")
        
    def test_copyfield(self):
        qp = qparser.QueryParser("a", None)
        qp.add_plugin(qparser.CopyFieldPlugin({"b": "c"}))
        self.assertEqual(unicode(qp.parse("hello b:matt")),
                         "(a:hello AND (b:matt OR c:matt))")
        
        qp = qparser.QueryParser("a", None)
        qp.add_plugin(qparser.CopyFieldPlugin({"b": "c"}, qparser.AndMaybeGroup))
        self.assertEqual(unicode(qp.parse("hello b:matt")),
                         "(a:hello AND (b:matt ANDMAYBE c:matt))")
        
        qp = qparser.QueryParser("a", None)
        qp.add_plugin(qparser.CopyFieldPlugin({"b": "c"}, qparser.RequireGroup))
        self.assertEqual(unicode(qp.parse("hello (there OR b:matt)")),
                         "(a:hello AND (a:there OR (b:matt REQUIRE c:matt)))")
        
        qp = qparser.QueryParser("a", None)
        qp.add_plugin(qparser.CopyFieldPlugin({"a": "c"}))
        self.assertEqual(unicode(qp.parse("hello there")),
                         "((a:hello OR c:hello) AND (a:there OR c:there))")
        
        qp = qparser.QueryParser("a", None)
        qp.add_plugin(qparser.CopyFieldPlugin({"b": "c"}, mirror=True))
        self.assertEqual(unicode(qp.parse("hello c:matt")),
                         "(a:hello AND (c:matt OR b:matt))")
        
        qp = qparser.QueryParser("a", None)
        qp.add_plugin(qparser.CopyFieldPlugin({"c": "a"}, mirror=True))
        self.assertEqual(unicode(qp.parse("hello c:matt")),
                         "((a:hello OR c:hello) AND (c:matt OR a:matt))")
        
        ana = analysis.RegexAnalyzer(r"\w+") | analysis.DoubleMetaphoneFilter()
        fmt = formats.Frequency(ana)
        schema = fields.Schema(name=fields.KEYWORD, name_phone=fields.FieldType(fmt, multitoken_query="or"))
        qp = qparser.QueryParser("name", schema)
        qp.add_plugin(qparser.CopyFieldPlugin({"name": "name_phone"}))
        self.assertEqual(unicode(qp.parse(u"spruce view")),
                         "((name:spruce OR name_phone:SPRS) AND (name:view OR name_phone:F OR name_phone:FF))")
        
    def test_gtlt(self):
        schema = fields.Schema(a=fields.KEYWORD, b=fields.NUMERIC,
                               c=fields.KEYWORD,
                               d=fields.NUMERIC(float), e=fields.DATETIME)
        qp = qparser.QueryParser("a", schema)
        qp.add_plugin(qparser.GtLtPlugin())
        qp.add_plugin(qparser.dateparse.DateParserPlugin())
        
        q = qp.parse(u"a:hello b:>100 c:<=z there")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(len(q), 4, unicode(q))
        self.assertEqual(q[0], query.Term("a", "hello"))
        self.assertEqual(q[1], query.NumericRange("b", 100, None, startexcl=True))
        self.assertEqual(q[2], query.TermRange("c", None, 'z'))
        self.assertEqual(q[3], query.Term("a", "there"))
        
        q = qp.parse(u"hello e:>'29 mar 2001' there")
        self.assertEqual(q.__class__, query.And)
        self.assertEqual(len(q), 3)
        self.assertEqual(q[0], query.Term("a", "hello"))
        # As of this writing, date ranges don't support startexcl/endexcl
        self.assertEqual(q[1], query.DateRange("e", datetime(2001, 3, 29, 0, 0), None))
        self.assertEqual(q[2], query.Term("a", "there"))
        
        q = qp.parse(u"a:> alfa c:<= bravo")
        self.assertEqual(unicode(q), "(a:a: AND a:alfa AND a:c: AND a:bravo)")
        
        qp.remove_plugin_class(qparser.FieldsPlugin)
        qp.remove_plugin_class(qparser.RangePlugin)
        q = qp.parse(u"hello a:>500 there")
        self.assertEqual(unicode(q), "(a:hello AND a:a: AND a:500 AND a:there)")
        
        


if __name__ == '__main__':
    unittest.main()








from __future__ import with_statement
import unittest

from whoosh import analysis, fields, formats, query, spans
from whoosh.filedb.filestore import RamStorage
from whoosh.query import And, Or, Term
from whoosh.util import permutations


class TestSpans(unittest.TestCase):
    domain = ("alfa", "bravo", "bravo", "charlie", "delta", "echo")
    
    def get_index(self):
        if hasattr(self, "_ix"):
            return self._ix
        
        ana = analysis.SimpleAnalyzer()
        charfield = fields.FieldType(format=formats.Characters(ana),
                                     scorable=True, stored=True)
        schema = fields.Schema(text=charfield)
        st = RamStorage()
        ix = st.create_index(schema)
        
        w = ix.writer()
        for ls in permutations(self.domain, 4):
            w.add_document(text=u" ".join(ls), _stored_text=ls)
        w.commit()
        
        self._ix = ix
        return ix
    
    def test_multimatcher(self):
        schema = fields.Schema(content=fields.TEXT(stored=True))
        ix = RamStorage().create_index(schema)
        
        domain = ("alfa", "bravo", "charlie", "delta")
        
        for _ in xrange(3):
            w = ix.writer()
            for ls in permutations(domain):
                w.add_document(content=u" ".join(ls))
            w.commit(merge=False)
        
        q = Term("content", "bravo")
        with ix.searcher() as s:
            m = q.matcher(s)
            while m.is_active():
                content = s.stored_fields(m.id())["content"].split()
                spans = m.spans()
                for span in spans:
                    self.assertEqual(content[span.start], "bravo")
                m.next()
    
    def test_excludematcher(self):
        schema = fields.Schema(content=fields.TEXT(stored=True))
        ix = RamStorage().create_index(schema)
        
        domain = ("alfa", "bravo", "charlie", "delta")
        
        for _ in xrange(3):
            w = ix.writer()
            for ls in permutations(domain):
                w.add_document(content=u" ".join(ls))
            w.commit(merge=False)
        
        w = ix.writer()
        w.delete_document(5)
        w.delete_document(10)
        w.delete_document(28)
        w.commit(merge=False)
        
        q = Term("content", "bravo")
        with ix.searcher() as s:
            m = q.matcher(s)
            while m.is_active():
                content = s.stored_fields(m.id())["content"].split()
                spans = m.spans()
                for span in spans:
                    self.assertEqual(content[span.start], "bravo")
                m.next()
        
    
    def test_span_term(self):
        ix = self.get_index()
        with ix.searcher() as s:
            alllists = [d["text"] for d in s.all_stored_fields()]
            
            for word in self.domain:
                q = Term("text", word)
                m = q.matcher(s)
                
                ids = set()
                while m.is_active():
                    id = m.id()
                    sps = m.spans()
                    ids.add(id)
                    original = list(s.stored_fields(id)["text"])
                    self.assertTrue(word in original, "%r not in %r" % (word, original))
                    
                    if word != "bravo":
                        self.assertEqual(len(sps), 1)
                    self.assertEqual(original.index(word), sps[0].start)
                    self.assertEqual(original.index(word), sps[0].end)
                    m.next()
            
                for i, ls in enumerate(alllists):
                    if word in ls:
                        self.assertTrue(i in ids)
                    else:
                        self.assertFalse(i in ids)
                    
    def test_span_first(self):
        ix = self.get_index()
        with ix.searcher() as s:
            for word in self.domain:
                q = spans.SpanFirst(Term("text", word))
                m = q.matcher(s)
                while m.is_active():
                    sps = m.spans()
                    original = s.stored_fields(m.id())["text"]
                    self.assertEqual(original[0], word)
                    self.assertEqual(len(sps), 1)
                    self.assertEqual(sps[0].start, 0)
                    self.assertEqual(sps[0].end, 0)
                    m.next()
                    
            q = spans.SpanFirst(Term("text", "bravo"), limit=1)
            m = q.matcher(s)
            while m.is_active():
                orig = s.stored_fields(m.id())["text"]
                for sp in m.spans():
                    self.assertEqual(orig[sp.start], "bravo")
                m.next()
            
    def test_span_near(self):
        ix = self.get_index()
        with ix.searcher() as s:
            def test(q):
                m = q.matcher(s)
                while m.is_active():
                    yield s.stored_fields(m.id())["text"], m.spans()
                    m.next()
                    
            for orig, sps in test(spans.SpanNear(Term("text", "alfa"), Term("text", "bravo"), ordered=True)):
                self.assertEqual(orig[sps[0].start], "alfa")
                self.assertEqual(orig[sps[0].end], "bravo")
                
            for orig, sps in test(spans.SpanNear(Term("text", "alfa"), Term("text", "bravo"), ordered=False)):
                first = orig[sps[0].start]
                second = orig[sps[0].end]
                self.assertTrue((first == "alfa" and second == "bravo")
                                or (first == "bravo" and second == "alfa"))
                
            for orig, sps in test(spans.SpanNear(Term("text", "bravo"), Term("text", "bravo"), ordered=True)):
                text = " ".join(orig)
                self.assertTrue(text.find("bravo bravo") > -1)
                
            q = spans.SpanNear(spans.SpanNear(Term("text", "alfa"), Term("text", "charlie")), Term("text", "echo"))
            for orig, sps in test(q):
                text = " ".join(orig)
                self.assertTrue(text.find("alfa charlie echo") > -1)
                
            q = spans.SpanNear(Or([Term("text", "alfa"), Term("text", "charlie")]), Term("text", "echo"), ordered=True)
            for orig, sps in test(q):
                text = " ".join(orig)
                self.assertTrue(text.find("alfa echo") > -1 or text.find("charlie echo") > -1)
    
    def test_near_unordered(self):
        schema = fields.Schema(text=fields.TEXT(stored=True))
        st = RamStorage()
        ix = st.create_index(schema)
        w = ix.writer()
        w.add_document(text=u"alfa bravo charlie delta echo")
        w.add_document(text=u"alfa bravo delta echo charlie")
        w.add_document(text=u"alfa charlie bravo delta echo")
        w.add_document(text=u"echo delta alfa foxtrot")
        w.commit()
        
        with ix.searcher() as s:
            q = spans.SpanNear(Term("text", "bravo"), Term("text", "charlie"), ordered=False)
            r = sorted(d["text"] for d in s.search(q))
            self.assertEqual(r, [u'alfa bravo charlie delta echo',
                                 u'alfa charlie bravo delta echo'])
        
    def test_span_near2(self):
        ana = analysis.SimpleAnalyzer()
        schema = fields.Schema(text=fields.TEXT(analyzer=ana, stored=True))
        st = RamStorage()
        ix = st.create_index(schema)
        w = ix.writer()
        w.add_document(text=u"The Lucene library is by Doug Cutting and Whoosh was made by Matt Chaput")
        w.commit()
        
        nq1 = spans.SpanNear(Term("text", "lucene"), Term("text", "doug"), slop=5)
        nq2 = spans.SpanNear(nq1, Term("text", "whoosh"), slop=4)
        
        with ix.searcher() as s:
            m = nq2.matcher(s)
            self.assertEqual(m.spans(), [spans.Span(1, 8)])
        
    def test_span_not(self):
        ix = self.get_index()
        with ix.searcher() as s:
            nq = spans.SpanNear(Term("text", "alfa"), Term("text", "charlie"), slop=2)
            bq = Term("text", "bravo")
            q = spans.SpanNot(nq, bq)
            m = q.matcher(s)
            while m.is_active():
                orig = list(s.stored_fields(m.id())["text"])
                i1 = orig.index("alfa")
                i2 = orig.index("charlie")
                dist = i2 - i1
                self.assertTrue(dist > 0 and dist < 3)
                if "bravo" in orig:
                    self.assertTrue(orig.index("bravo") != i1 + 1)
                m.next()
            
    def test_span_or(self):
        ix = self.get_index()
        with ix.searcher() as s:
            nq = spans.SpanNear(Term("text", "alfa"), Term("text", "charlie"), slop=2)
            bq = Term("text", "bravo")
            q = spans.SpanOr([nq, bq])
            m = q.matcher(s)
            while m.is_active():
                orig = s.stored_fields(m.id())["text"]
                self.assertTrue(("alfa" in orig and "charlie" in orig) or "bravo" in orig)
                m.next()

    def test_span_contains(self):
        ix = self.get_index()
        with ix.searcher() as s:
            nq = spans.SpanNear(Term("text", "alfa"), Term("text", "charlie"), slop=3)
            cq = spans.SpanContains(nq, Term("text", "echo"))
            
            m = cq.matcher(s)
            ls = []
            while m.is_active():
                orig = s.stored_fields(m.id())["text"]
                ls.append(" ".join(orig))
                m.next()
            ls.sort()
            self.assertEqual(ls, ['alfa bravo echo charlie', 'alfa bravo echo charlie',
                                  'alfa delta echo charlie', 'alfa echo bravo charlie',
                                  'alfa echo bravo charlie', 'alfa echo charlie bravo',
                                  'alfa echo charlie bravo', 'alfa echo charlie delta',
                                  'alfa echo delta charlie', 'bravo alfa echo charlie',
                                  'bravo alfa echo charlie', 'delta alfa echo charlie'])

    def test_span_before(self):
        ix = self.get_index()
        with ix.searcher() as s:
            bq = spans.SpanBefore(Term("text", "alfa"), Term("text", "charlie"))
            m = bq.matcher(s)
            while m.is_active():
                orig = list(s.stored_fields(m.id())["text"])
                self.assertTrue("alfa" in orig)
                self.assertTrue("charlie" in orig)
                self.assertTrue(orig.index("alfa") < orig.index("charlie"))
                m.next()

    def test_span_condition(self):
        ix = self.get_index()
        with ix.searcher() as s:
            sc = spans.SpanCondition(Term("text", "alfa"), Term("text", "charlie"))
            m = sc.matcher(s)
            while m.is_active():
                orig = list(s.stored_fields(m.id())["text"])
                self.assertTrue("alfa" in orig)
                self.assertTrue("charlie" in orig)
                for span in m.spans():
                    self.assertEqual(orig[span.start], "alfa")
                m.next()

    def test_regular_or(self):
        ix = self.get_index()
        with ix.searcher() as s:
            oq = query.Or([query.Term("text", "bravo"), query.Term("text", "alfa")])
            m = oq.matcher(s)
            while m.is_active():
                orig = s.stored_fields(m.id())["text"]
                for span in m.spans():
                    v = orig[span.start]
                    self.assertTrue(v == "bravo" or v == "alfa")
                m.next()
            
    def test_regular_and(self):
        ix = self.get_index()
        with ix.searcher() as s:
            aq = query.And([query.Term("text", "bravo"), query.Term("text", "alfa")])
            m = aq.matcher(s)
            while m.is_active():
                orig = s.stored_fields(m.id())["text"]
                for span in m.spans():
                    v = orig[span.start]
                    self.assertTrue(v == "bravo" or v == "alfa")
                m.next()

    def test_span_characters(self):
        ix = self.get_index()
        with ix.searcher() as s:
            pq = query.Phrase("text", ["bravo", "echo"])
            m = pq.matcher(s)
            while m.is_active():
                orig = " ".join(s.stored_fields(m.id())["text"])
                for span in m.spans():
                    startchar, endchar = span.startchar, span.endchar
                    self.assertEqual(orig[startchar:endchar], "bravo echo")
                m.next()

    


            
if __name__ == '__main__':
    unittest.main()














#===============================================================================
# Copyright 2010 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""
This module contains plugins for the query parser. Most of the functionality
of the default query parser is actually provided by plugins.
"""

import re

from whoosh.qparser.syntax import *
from whoosh.qparser.common import get_single_text, rcompile, QueryParserError


class Plugin(object):
    """Base class for parser plugins.
    """
            
    def tokens(self, parser):
        """Returns a list of ``(token_class, priority)`` tuples to add to the
        syntax the parser understands.
        """
        
        return ()
    
    def filters(self, parser):
        """Returns a list of ``(filter_function, priority)`` tuples to add to
        parser.
        """
        
        return ()
    

class RangePlugin(Plugin):
    """Adds the ability to specify term ranges.
    
    This plugin has no configuration.
    
    This plugin is included in the default parser configuration.
    """
    
    def tokens(self, parser):
        return ((RangePlugin.Range, 1), )
    
    class Range(Token):
        expr = re.compile(r"""
        (?P<open>\{|\[)               # Open paren
        (?P<start>
            ('[^']*?'\s+)             # single-quoted 
            |                         # or
            (.+?(?=[Tt][Oo]))         # everything until "to"
        )?
        [Tt][Oo]                      # "to"
        (?P<end>
            (\s+'[^']*?')             # single-quoted
            |                         # or
            ((.+?)(?=]|}))            # everything until "]" or "}"
        )?
        (?P<close>}|])                # Close paren
        """, re.UNICODE | re.VERBOSE)
        
        def __init__(self, start, end, startexcl, endexcl, fieldname=None, boost=1.0):
            self.fieldname = fieldname
            self.start = start
            self.end = end
            self.startexcl = startexcl
            self.endexcl = endexcl
            self.boost = boost
        
        def set_boost(self, b):
            return self.__class__(self.start, self.end, self.startexcl,
                                  self.endexcl, fieldname=self.fieldname,
                                  boost=b)
        
        def set_fieldname(self, name, force=False):
            if force or self.fieldname is None:
                return self.__class__(self.start, self.end, self.startexcl,
                                      self.endexcl, fieldname=name,
                                      boost=self.boost)
            else:
                return self
        
        def __repr__(self):
            r = "%s:(%r, %r, %s, %s)" % (self.fieldname, self.start, self.end,
                                         self.startexcl, self.endexcl)
            if self.boost != 1.0:
                r += "^%s" % self.boost
            return r
        
        @classmethod
        def create(cls, parser, match):
            start = match.group("start")
            end = match.group("end")
            if start:
                start = start.rstrip()
                if start.startswith("'") and start.endswith("'"):
                    start = start[1:-1]
            if end:
                end = end.lstrip()
                if end.startswith("'") and end.endswith("'"):
                    end = end[1:-1]
            
            return cls(start, end, startexcl=match.group("open") == "{",
                       endexcl=match.group("close") == "}")
            
        def query(self, parser):
            fieldname = self.fieldname or parser.fieldname
            start, end = self.start, self.end
            if parser.schema and fieldname in parser.schema:
                field = parser.schema[fieldname]
                
                if field.self_parsing():
                    try:
                        rangeq = field.parse_range(fieldname, start, end,
                                                   self.startexcl, self.endexcl,
                                                   boost=self.boost)
                        if rangeq is not None:
                            return rangeq
                    except QueryParserError, e:
                        return query.NullQuery
                
                if start:
                    start = get_single_text(field, start, tokenize=False,
                                            removestops=False)
                if end:
                    end = get_single_text(field, end, tokenize=False,
                                          removestops=False)
            
            return query.TermRange(fieldname, start, end, self.startexcl,
                                   self.endexcl, boost=self.boost)
            

class PhrasePlugin(Plugin):
    """Adds the ability to specify phrase queries inside double quotes.
    
    This plugin has no configuration.
    
    This plugin is included in the default parser configuration.
    """
    
    def tokens(self, parser):
        return ((PhrasePlugin.Quotes, 0), )
    
    class Quotes(BasicSyntax):
        expr = rcompile('"(.*?)"')
        
        def __init__(self, text, fieldname=None, boost=1.0, slop=1):
            super(PhrasePlugin.Quotes, self).__init__(text, fieldname=fieldname,
                                                      boost=boost)
            self.slop = slop
        
        def __repr__(self):
            r = "%s:q(%r)" % (self.fieldname, self.text)
            if self.boost != 1.0:
                r += "^%s" % self.boost
            return r
        
        @classmethod
        def create(cls, parser, match):
            slop = 1
            #if match.group(5):
            #    try:
            #        slop = int(match.group(5))
            #    except ValueError:
            #        pass
            return cls(match.group(1), slop=slop)
        
        def query(self, parser):
            fieldname = self.fieldname or parser.fieldname
            if parser.schema and fieldname in parser.schema:
                field = parser.schema[fieldname]
                #if field.self_parsing():
                #    return field.parse_query(fieldname, self.text, boost=self.boost)
                #else:
                words = list(field.process_text(self.text, mode="query")) 
            else:
                words = self.text.split(" ")
            
            return parser.phraseclass(fieldname, words, boost=self.boost,
                                      slop=self.slop)


class SingleQuotesPlugin(Plugin):
    """Adds the ability to specify single "terms" containing spaces by
    enclosing them in single quotes.
    
    This plugin has no configuration.
    
    This plugin is included in the default parser configuration.
    """
     
    def tokens(self, parser):
        return ((SingleQuotesPlugin.SingleQuotes, 0), )
    
    class SingleQuotes(Token):
        expr = rcompile(r"(^|(?<=\W))'(.*?)'(?=\s|\]|[)}]|$)")
        
        @classmethod
        def create(cls, parser, match):
            return Word(match.group(2))


class PrefixPlugin(Plugin):
    """Adds the ability to specify prefix queries by ending a term with an
    asterisk. This plugin is useful if you want the user to be able to create
    prefix but not wildcard queries (for performance reasons). If you are
    including the wildcard plugin, you should not include this plugin as well.
    """
    
    def tokens(self, parser):
        return ((PrefixPlugin.Prefix, 0), )
    
    class Prefix(BasicSyntax):
        expr = rcompile("[^ \t\r\n*]+\\*(?= |$|\\))")
        qclass = query.Prefix
        
        def __repr__(self):
            r = "%s:pre(%r)" % (self.fieldname, self.text)
            if self.boost != 1.0:
                r += "^%s" % self.boost
            return r
        
        @classmethod
        def create(cls, parser, match):
            return cls(match.group(0)[:-1])
        

class WildcardPlugin(Plugin):
    """Adds the ability to specify wildcard queries by using asterisk and
    question mark characters in terms. Note that these types can be very
    performance and memory intensive. You may consider not including this
    type of query.
    
    This plugin is included in the default parser configuration.
    """
    
    def tokens(self, parser):
        return ((WildcardPlugin.Wild, 1), )
    
    class Wild(BasicSyntax):
        # \u055E = Armenian question mark
        # \u061F = Arabic question mark
        # \u1367 = Ethiopic question mark
        expr = rcompile(u"[^ \t\r\n*?\u055E\u061F\u1367]*[*?\u055E\u061F\u1367]\\S*")
        qclass = query.Wildcard
        
        def __repr__(self):
            r = "%s:wild(%r)" % (self.fieldname, self.text)
            if self.boost != 1.0:
                r += "^%s" % self.boost
            return r
        
        @classmethod
        def create(cls, parser, match):
            return cls(match.group(0))
        

class WhitespacePlugin(Plugin):
    """Parses whitespace between words in the query string. You should always
    include this plugin.
    
    This plugin is always automatically included by the QueryParser.
    """
    
    def __init__(self, tokenclass=White):
        self.tokenclass = tokenclass
    
    def tokens(self, parser):
        return ((self.tokenclass, 100), )
    
    def filters(self, parser):
        return ((self.do_whitespace, 500), )
    
    def do_whitespace(self, parser, stream):
        newstream = stream.empty()
        for t in stream:
            if isinstance(t, Group):
                newstream.append(self.do_whitespace(parser, t))
            elif not isinstance(t, self.tokenclass):
                newstream.append(t)
        return newstream


class GroupPlugin(Plugin):
    """Adds the ability to group clauses using parentheses.
    
    This plugin is included in the default parser configuration.
    """
    
    def tokens(self, parser):
        return ((GroupPlugin.Open, 0), (GroupPlugin.Close, 0))
    
    def filters(self, parser):
        # This should basically be the first plugin to run
        return ((GroupPlugin.do_groups, 0), )
    
    @staticmethod
    def do_groups(parser, stream):
        stack = [parser.group()]
        for t in stream:
            if isinstance(t, GroupPlugin.Open):
                stack.append(parser.group())
            elif isinstance(t, GroupPlugin.Close):
                if len(stack) > 1:
                    last = stack.pop()
                    stack[-1].append(last)
            else:
                stack[-1].append(t)
        
        top = stack[0]
        if len(stack) > 1:
            for ls in stack[1:]:
                top.extend(ls)
        
        if len(top) == 1 and isinstance(top[0], Group):
            top = top[0].set_boost(top.boost)
        
        return top
    
    class Open(Singleton):
        expr = rcompile("\\(")
        
    class Close(Singleton):
        expr = rcompile("\\)")


class FieldsPlugin(Plugin):
    """Adds the ability to specify the field of a clause using a colon.
    
    This plugin is included in the default parser configuration.
    """
    
    def tokens(self, parser):
        return ((FieldsPlugin.Field, 0), )
    
    def filters(self, parser):
        return ((FieldsPlugin.do_fieldnames, 100), )

    @staticmethod
    def do_fieldnames(parser, stream):
        fieldtoken = FieldsPlugin.Field
        newstream = stream.empty()
        
        i = len(stream)
        # Iterate backwards through the stream, looking for field-able objects
        # with field tokens in front of them
        while i > 0:
            i -= 1
            t = stream[i]
            
            if isinstance(t, fieldtoken):
                # If this we see a field token in the stream, it means it
                # wasn't in front of a field-able object, so convert it into a
                # Word token
                t = Word(t.original)
            elif isinstance(t, Group):
                t = FieldsPlugin.do_fieldnames(parser, t)
            
            # If this is a field-able object (not whitespace or a field token)
            # and it has a field token in front of it, apply the field token
            if (i > 0 and not isinstance(t, (White, fieldtoken))
                and isinstance(stream[i - 1], fieldtoken)):
                # Set the field name for this object from the field token
                t = t.set_fieldname(stream[i - 1].fieldname)
                # Skip past the field token
                i -= 1
            
            newstream.append(t)

        newstream.reverse()
        return newstream
    
    class Field(Token):
        expr = rcompile(r"(?P<fieldname>\w[\w\d]*):")
        
        def __init__(self, fieldname, original):
            self.fieldname = fieldname
            self.original = original
        
        def __repr__(self):
            return "<%s:>" % self.fieldname
        
        def set_fieldname(self, fieldname, force=False):
            return self.__class__(fieldname, self.original)
        
        @classmethod
        def create(cls, parser, match):
            fieldname = match.group("fieldname")
            if not parser.schema or fieldname == "*" or (fieldname in parser.schema):
                return cls(fieldname, match.group(0))
    

class OperatorsPlugin(Plugin):
    """By default, adds the AND, OR, ANDNOT, ANDMAYBE, and NOT operators to
    the parser syntax. This plugin scans the token stream for subclasses of
    :class:`Operator` and calls their :meth:`Operator.make_group` methods
    to allow them to manipulate the stream.
    
    There are two levels of configuration available.
    
    The first level is to change the regular expressions of the default
    operators, using the ``And``, ``Or``, ``AndNot``, ``AndMaybe``, and/or
    ``Not`` keyword arguments. The keyword value can be a pattern string or
    a compiled expression, or None to remove the operator::
    
        qp = qparser.QueryParser("content", schema)
        cp = qparser.OperatorsPlugin(And="&", Or="\\|", AndNot="&!", AndMaybe="&~", Not=None)
        qp.replace_plugin(cp)
    
    You can also specify a list of ``(Operator, priority)`` pairs as the first
    argument to the initializer. For example, assume you have created an
    :class:`InfixOperator` subclass to implement a "before" operator. To add
    this to the operators plugin with a priority of -5, you would do this::
    
        additional = [(MyBefore(), -5)]
        cp = qparser.OperatorsPlugin(additional)
    
    Not that the list of operators you specify with the first argument is IN
    ADDITION TO the defaults. To turn off one of the default operators, you
    can pass None to the corresponding keyword argument::
        
        cp = qparser.OperatorsPlugin([(MyAnd(), 0)], And=None)
        
    If you want ONLY your list operators and none of the default operators, use
    the ``clean`` keyword argument::
    
        cp = qparser.OperatorsPlugin([(MyAnd(), 0)], clean=True)
                                     
    This class replaces the ``CompoundsPlugin``. ``qparser.CompoundsPlugin`` is
    now an alias for this class.
    """
    
    def __init__(self, ops=None, And=r"\sAND\s", Or=r"\sOR\s",
                 AndNot=r"\sANDNOT\s", AndMaybe=r"\sANDMAYBE\s",
                 Not=r"(^|(?<= ))NOT\s", Require=r"(^|(?<= ))REQUIRE\s",
                 clean=False):
        if isinstance(ops, tuple):
            ops = list(ops)
        if not ops:
            ops = []
        
        if not clean:
            if Not:
                ops.append((PrefixOperator(Not, NotGroup), 0))
            if And:
                ops.append((InfixOperator(And, AndGroup), 0))
            if Or:
                ops.append((InfixOperator(Or, OrGroup), 0))
            if AndNot:
                ops.append((InfixOperator(AndNot, AndNotGroup), -5))
            if AndMaybe:
                ops.append((InfixOperator(AndMaybe, AndMaybeGroup), -5))
            if Require:
                ops.append((InfixOperator(Require, RequireGroup), 0))
        
        self.ops = ops
    
    def tokens(self, parser):
        return self.ops
    
    def filters(self, parser):
        return ((self.do_operators, 600), )
    
    def do_operators(self, parser, stream, level=0):
        for op, _ in self.ops:
            if op.left_assoc:
                i = 0
                while i < len(stream):
                    t = stream[i]
                    if t is op:
                        i = t.make_group(parser, stream, i)
                    else:
                        i += 1
            else:
                i = len(stream) - 1
                while i >= 0:
                    t = stream[i]
                    if t is op:
                        i = t.make_group(parser, stream, i)
                    i -= 1
        
        #print " " * level, ">stream=", stream
        newstream = stream.empty()
        for t in stream:
            if isinstance(t, Group):
                t = self.do_operators(parser, t, level + 1)
            newstream.append(t)
        
        #print " " * level, "<stream=", newstream
        return newstream

CompoundsPlugin = OperatorsPlugin


class NotPlugin(Plugin):
    """This plugin is deprecated, its functionality is now provided by the
    :class:`OperatorsPlugin`.
    """
    
    def __init__(self, token="(^|(?<= ))NOT "):
        class Not(Singleton):
            expr = rcompile(token)
        
        self.Not = Not
    
    def tokens(self, parser):
        return ((self.Not, 0), )
    
    def filters(self, parser):
        return ((self.do_not, 800), )
    
    def do_not(self, parser, stream):
        newstream = stream.empty()
        notnext = False
        for t in stream:
            if isinstance(t, self.Not):
                notnext = True
                continue
            
            if isinstance(t, Group):
                t = self.do_not(parser, t)
            
            if notnext:
                t = NotGroup([t])
            
            newstream.append(t)
            notnext = False
            
        return newstream
 

class BoostPlugin(Plugin):
    """Adds the ability to boost clauses of the query using the circumflex.
    
    This plugin is included in the default parser configuration.
    """
    
    def tokens(self, parser):
        return ((BoostPlugin.Boost, 0), )
    
    def filters(self, parser):
        return ((BoostPlugin.clean_boost, 0), (BoostPlugin.do_boost, 700))

    @staticmethod
    def clean_boost(parser, stream):
        newstream = stream.empty()
        for i, t in enumerate(stream):
            if isinstance(t, BoostPlugin.Boost):
                if i == 0 or isinstance(stream[i - 1], (BoostPlugin.Boost, White)):
                    t = Word(t.original)
            newstream.append(t)
        return newstream

    @staticmethod
    def do_boost(parser, stream):
        newstream = stream.empty()
        
        for t in stream:
            if isinstance(t, Group):
                newstream.append(BoostPlugin.do_boost(parser, t))
                
            elif isinstance(t, BoostPlugin.Boost):
                if newstream:
                    newstream.append(newstream.pop().set_boost(t.boost))
                
            else:
                newstream.append(t)
        
        return newstream
    
    class Boost(Token):
        expr = rcompile("\\^([0-9]+(.[0-9]+)?)($|(?=[ \t\r\n]))")
        
        def __init__(self, original, boost):
            self.original = original
            self.boost = boost
        
        def __repr__(self):
            return "<^%s>" % self.boost
        
        @classmethod
        def create(cls, parser, match):
            try:
                return cls(match.group(0), float(match.group(1)))
            except ValueError:
                return Word(match.group(0))
    

class PlusMinusPlugin(Plugin):
    """Adds the ability to use + and - in a flat OR query to specify required
    and prohibited terms.
    
    This is the basis for the parser configuration returned by
    ``SimpleParser()``.
    """
    
    def tokens(self, parser):
        return ((PlusMinusPlugin.Plus, 0), (PlusMinusPlugin.Minus, 0))
    
    def filters(self, parser):
        return ((PlusMinusPlugin.do_plusminus, 510), )
    
    @staticmethod
    def do_plusminus(parser, stream):
        required = AndGroup()
        optional = OrGroup()
        prohibited = OrGroup()
        
        nextlist = optional
        for t in stream:
            if isinstance(t, PlusMinusPlugin.Plus):
                nextlist = required
            elif isinstance(t, PlusMinusPlugin.Minus):
                nextlist = prohibited
            else:
                nextlist.append(t)
                nextlist = optional
        
        r = optional
        if required:
            r = AndMaybeGroup([required, optional])
        if prohibited:
            r = AndNotGroup([r, prohibited])
        return r
    
    class Plus(Singleton):
        expr = rcompile("\\+")
        
    class Minus(Singleton):
        expr = rcompile("-")


class MultifieldPlugin(Plugin):
    """Converts any unfielded terms into OR clauses that search for the
    term in a specified list of fields.
    """
    
    def __init__(self, fieldnames, fieldboosts=None):
        """
        :param fieldnames: a list of fields to search.
        :param fieldboosts: an optional dictionary mapping field names to
            a boost to use for that field.
        """
        
        self.fieldnames = fieldnames
        self.boosts = fieldboosts or {}
    
    def filters(self, parser):
        return ((self.do_multifield, 110), )
    
    def do_multifield(self, parser, stream):
        newstream = stream.empty()
        for t in stream:
            if isinstance(t, BasicSyntax) and t.fieldname is None:
                t = OrGroup([t.set_fieldname(fn).set_boost(self.boosts.get(fn, 1.0))
                             for fn in self.fieldnames])
            newstream.append(t)
        return newstream
        

class DisMaxPlugin(Plugin):
    """Converts any unfielded terms into DisjunctionMax clauses that search
    for the term in a specified list of fields.
    """
    
    def __init__(self, fieldboosts, tiebreak=0.0):
        """
        :param fieldboosts: a dictionary mapping field names to a boost to use
            for that in the DisjuctionMax query.
        """
        
        self.fieldboosts = fieldboosts.items()
        self.tiebreak = tiebreak
    
    def filters(self, parser):
        return ((self.do_dismax, 110), )
    
    def do_dismax(self, parser, stream):
        newstream = stream.empty()
        for t in stream:
            if isinstance(t, BasicSyntax) and t.fieldname is None:
                t = DisMaxGroup([t.set_fieldname(fn).set_boost(b)
                                 for fn, b in self.fieldboosts],
                                 tiebreak=self.tiebreak)
            newstream.append(t)
        return newstream


class FieldAliasPlugin(Plugin):
    """Adds the ability to use "aliases" of fields in the query string.
    
    >>> # Allow users to use 'body' or 'text' to refer to the 'content' field
    >>> parser.add_plugin(FieldAliasPlugin({"content": ("body", "text")}))
    >>> parser.parse("text:hello")
    Term("content", "hello")
    """
    
    def __init__(self, fieldmap):
        """
        :param fieldmap: a dictionary mapping fieldnames to a list of
            aliases for the field.
        """
        
        self.fieldmap = fieldmap
        self.reverse = {}
        for key, values in fieldmap.iteritems():
            for value in values:
                self.reverse[value] = key
        
    def filters(self, parser):
        return ((self.do_aliases, 90), )
    
    def do_aliases(self, parser, stream):
        newstream = stream.empty()
        for t in stream:
            if (not isinstance(t, Group)
                  and t.fieldname is not None
                  and t.fieldname in self.reverse):
                    t = t.set_fieldname(self.reverse[t.fieldname])
            newstream.append(t)
        return newstream


class CopyFieldPlugin(Plugin):
    """Looks for basic syntax tokens (terms, prefixes, wildcards, phrases, etc.)
    occurring in a certain field and replaces it with a group (by default OR)
    containing the original token and the token copied to a new field.
    
    For example, the query::
    
        hello name:matt
        
    could be automatically converted by ``CopyFieldPlugin({"name", "author"})``
    to::
    
        hello (name:matt OR author:matt)
    
    This is useful where one field was indexed with a differently-analyzed copy
    of another, and you want the query to search both fields.
    """
    
    def __init__(self, map, group=OrGroup, mirror=False):
        """
        :param map: a dictionary mapping names of fields to copy to the
            names of the destination fields.
        :param group: the type of group to create in place of the original
            token.
        :param two_way: if True, the plugin copies both ways, so if the user
            specifies a query in the 'toname' field, it will be copied to
            the 'fromname' field.
        """
        self.map = map
        self.group = group
        self.mirror = mirror
        
    def filters(self, parser):
        return ((self.do_copyfield, 109), )
    
    def do_copyfield(self, parser, stream):
        mirror = self.mirror
        map = self.map
        if mirror:
            # Add in reversed mappings
            map.update(dict((v, k) for k, v in map.iteritems()))
        
        newstream = stream.empty()
        for t in stream:
            if isinstance(t, Group):
                t = self.do_copyfield(parser, t)
            elif isinstance(t, BasicSyntax):
                toname = None
                if t.fieldname in map:
                    toname = map[t.fieldname]
                elif t.fieldname is None and parser.fieldname in map:
                    toname = map[parser.fieldname]
                
                if toname:
                    # Replace the syntax object with a group containing the
                    # original object and one with the field changed
                    t = self.group([t, t.set_fieldname(toname, force=True)])
            newstream.append(t)
        return newstream


class GtLtPlugin(Plugin):
    """Allows the user to use greater than/less than symbols to create range
    queries::
    
        a:>100 b:<=z c:>=-1.4 d:<mz
        
    This is the equivalent of::
    
        a:{100 to] b:[to z] c:[-1.4 to] d:[to mz}
        
    The plugin recognizes ``>``, ``<``, ``>=``, ``<=``, ``=>``, and ``=<``
    after a field specifier. The field specifier is required. You cannot do the
    following::
    
        >100
        
    This plugin requires the FieldsPlugin and RangePlugin to work.
    """
    
    def __init__(self, expr=r"(?P<rel>(<=|>=|<|>|=<|=>))"):
        """
        :param expr: a regular expression that must capture a "rel" group
            (which contains <, >, >=, <=, =>, or =<)
        """
        
        self.expr = rcompile(expr)
    
    def tokens(self, parser):
        # Create a dynamic subclass of GtLtToken and give it the configured
        # regular expression
        tkclass = type("DynamicGtLtToken", (GtLtPlugin.GtLtToken, ),
                       {"expr": self.expr})
        
        return ((tkclass, 0), )
    
    def filters(self, parser):
        # Run before the fieldnames filter
        return ((self.do_gtlt, 99), )
    
    def make_range(self, text, rel):
        if rel == "<":
            return RangePlugin.Range(None, text, False, True)
        elif rel == ">":
            return RangePlugin.Range(text, None, True, False)
        elif rel == "<=" or rel == "=<":
            return RangePlugin.Range(None, text, False, False)
        elif rel == ">=" or rel == "=>":
            return RangePlugin.Range(text, None, False, False)
    
    def do_gtlt(self, parser, stream):
        # Look for GtLtTokens in the stream and
        # - replace it with a Field token
        # - if the next token is a Word, replace it with a Range based on the
        #   GtLtToken
        
        gtlttoken = GtLtPlugin.GtLtToken
        
        newstream = stream.empty()
        prev = None
        for t in stream:
            if isinstance(t, gtlttoken):
                if not isinstance(prev, FieldsPlugin.Field):
                    prev = None
                    continue
            elif isinstance(t, Word) and isinstance(prev, gtlttoken):
                t = self.make_range(t.text, prev.rel)
            
            if not isinstance(t, gtlttoken):
                newstream.append(t)
            prev = t
        
        return newstream
    
    class GtLtToken(Token):
        def __init__(self, rel):
            self.rel = rel
        
        def __repr__(self):
            return "{%s}" % (self.rel)
        
        @classmethod
        def create(cls, parser, match):
            return cls(match.group("rel"))










import unittest

import os.path, shutil, tempfile
from random import randint, shuffle

from whoosh.filedb.filestore import FileStorage
from whoosh.filedb.filetables import HashWriter, HashReader, dump_hash

class Test(unittest.TestCase):
    def test_bigtable(self):
        dir = tempfile.mkdtemp(prefix="bigtable", suffix=".tmpix")
        st = FileStorage(dir)
        
        def randstring(min, max):
            return "".join(chr(randint(1, 255))
                           for _ in xrange(randint(min, max)))
        
        count = 100000
        samp = dict((randstring(1,50), randstring(1,50))
                    for _ in xrange(count))
        
        fhw = HashWriter(st.create_file("big.hsh"))
        fhw.add_all(samp.iteritems())
        fhw.close()
        
        fhr = HashReader(st.open_file("big.hsh"))
        keys = samp.keys()
        shuffle(keys)
        for key in keys:
            self.assertEqual(samp[key], fhr[key])
        
        set1 = set(samp.iteritems())
        set2 = set(fhr.items())
        self.assertEqual(set1, set2)
        
        fhr.close()
        shutil.rmtree(dir)


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



import unittest

import random, shutil, tempfile, threading, time

from whoosh import fields, index, query


class TestThreading(unittest.TestCase):
    def test_readwrite(self):
        dir = tempfile.mkdtemp(prefix="threading", suffix=".tmpix")
        
        ixname = "threading"
        schema = fields.Schema(id=fields.ID(stored=True), content=fields.TEXT)
        
        domain = ("alfa", "bravo", "charlie", "delta", "echo", "foxtrot",
                  "golf", "hotel", "india", "juliet", "kilo", "lima", "mike",
                  "november", "oscar", "papa", "quebec", "romeo", "sierra",
                  "tango", "uniform", "victor", "whiskey", "xray", "yankee",
                  "zulu")
        
        class WriterThread(threading.Thread):
            def run(self):
                ix = index.create_in(dir, schema, indexname=ixname)
                num = 0
                
                for i in xrange(50):
                    print i
                    w = ix.writer()
                    for _ in xrange(random.randint(1, 100)):
                        content = u" ".join(random.sample(domain, random.randint(5, 20)))
                        w.add_document(id=unicode(num), content=content)
                        num += 1
                    w.commit()
                    
                    time.sleep(0.1)
        
        class SearcherThread(threading.Thread):
            def run(self):
                print self.name + " starting"
                for _ in xrange(10):
                    ix = index.open_dir(dir, indexname=ixname)
                    s = ix.searcher()
                    q = query.Term("content", random.choice(domain))
                    s.search(q, limit=10)
                    s.close()
                    ix.close()
                    time.sleep(0.1)
                print self.name + " done"
        
        wt = WriterThread()
        wt.start()
        time.sleep(0.5)
        for _ in xrange(20):
            SearcherThread().start()
            time.sleep(0.5)
        wt.join()
        
        shutil.rmtree(dir)

if __name__ == '__main__':
    unittest.main()



#===============================================================================
# Copyright 2007 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""This module contains objects that query the search index. These query
objects are composable to form complex query trees.
"""

from __future__ import division

import copy
import fnmatch
import operator
import re
from array import array

from whoosh.lang.morph_en import variations
from whoosh.matching import (AndMaybeMatcher, DisjunctionMaxMatcher,
                             ListMatcher, IntersectionMatcher, InverseMatcher,
                             NullMatcher, RequireMatcher, UnionMatcher,
                             WrappingMatcher, ConstantScoreMatcher,
                             AndNotMatcher)
from whoosh.reading import TermNotFound
from whoosh.support.levenshtein import relative
from whoosh.support.times import datetime_to_long
from whoosh.util import make_binary_tree


# Exceptions

class QueryError(Exception):
    """Error encountered while running a query.
    """
    pass


# Utility classes

class Lowest(object):
    "A value that is always compares lower than any other object except itself."
    
    def __cmp__(self, other):
        if other.__class__ is Lowest:
            return 0
        return -1
Lowest = Lowest()

class Highest(object):
    "A value that is always compares higher than any other object except itself."
    
    def __cmp__(self, other):
        if other.__class__ is Highest:
            return 0
        return 1
Highest = Highest()


# Base classes

class Query(object):
    """Abstract base class for all queries.
    
    Note that this base class implements __or__, __and__, and __sub__ to allow
    slightly more convenient composition of query objects::
    
        >>> Term("content", u"a") | Term("content", u"b")
        Or([Term("content", u"a"), Term("content", u"b")])
        
        >>> Term("content", u"a") & Term("content", u"b")
        And([Term("content", u"a"), Term("content", u"b")])
        
        >>> Term("content", u"a") - Term("content", u"b")
        And([Term("content", u"a"), Not(Term("content", u"b"))])
    """

    def __or__(self, query):
        """Allows you to use | between query objects to wrap them in an Or
        query.
        """
        return Or([self, query]).normalize()

    def __and__(self, query):
        """Allows you to use & between query objects to wrap them in an And
        query.
        """
        return And([self, query]).normalize()

    def __sub__(self, query):
        """Allows you to use - between query objects to add the right-hand
        query as a "NOT" query.
        """

        return And([self, Not(query)]).normalize()

    def __hash__(self):
        raise NotImplementedError

    def is_leaf(self):
        """Returns True if this is a leaf node in the query tree, or False if
        this query has sub-queries.
        """
        
        return True

    def apply(self, fn):
        """If this query has children, calls the given function on each child
        and returns a new copy of this node with the new children returned by
        the function. If this is a leaf node, simply returns this object.
        
        This is useful for writing functions that transform a query tree. For
        example, this function changes all Term objects in a query tree into
        Variations objects::
        
            def term2var(q):
                if isinstance(q, Term):
                    return Variations(q.fieldname, q.text)
                else:
                    return q.apply(term2var)
        
            q = And([Term("f", "alfa"),
                     Or([Term("f", "bravo"),
                         Not(Term("f", "charlie"))])])
            q = term2var(q)
            
        Note that this method does not automatically create copies of nodes.
        To avoid modifying the original tree, your function should call the
        :meth:`Query.copy` method on nodes before changing their attributes.
        """
        
        return self

    def accept(self, fn):
        """Applies the given function to this query's subqueries (if any) and
        then to this query itself::
        
            def boost_phrases(q):
                if isintance(q, Phrase):
                    q.boost *= 2.0
                return q
            
            myquery = myquery.accept(boost_phrases)
        
        This method automatically creates copies of the nodes in the original
        tree before passing them to your function, so your function can change
        attributes on nodes without altering the original tree.
        
        This method is less flexible than using :meth:`Query.apply` (in fact
        it's implemented using that method) but is often more straightforward.
        """
        
        def fn_wrapper(q):
            q = q.apply(fn_wrapper)
            return fn(q)
        
        return fn_wrapper(self)

    def copy(self):
        """Returns a copy of this query tree.
        """
        
        if self.is_leaf():
            return copy.copy(self)
        else:
            return self.apply(operator.methodcaller("copy"))

    def replace(self, oldtext, newtext):
        """Returns a copy of this query with oldtext replaced by newtext (if
        oldtext was anywhere in this query).
        
        Note that this returns a *new* query with the given text replaced. It
        *does not* modify the original query "in place".
        """
        
        if self.is_leaf():
            return self.copy()
        else:
            return self.apply(operator.methodcaller("replace", oldtext, newtext))

    def all_terms(self, termset=None, phrases=True):
        """Returns a set of all terms in this query tree.
        
        This method simply operates on the query itself, without reference to
        an index (unlike existing_terms()), so it will *not* add terms that
        require an index to compute, such as Prefix and Wildcard.
        
        >>> q = And([Term("content", u"render"), Term("path", u"/a/b")])
        >>> q.all_terms()
        set([("content", u"render"), ("path", u"/a/b")])
        
        :param phrases: Whether to add words found in Phrase queries.
        :rtype: set
        """

        if termset is None:
            termset = set()
        self._all_terms(termset, phrases=phrases)
        return termset

    def _all_terms(self, *args, **kwargs):
        # To be implemented in sub-classes
        return

    def existing_terms(self, ixreader, termset=None, reverse=False,
                       phrases=True):
        """Returns a set of all terms in this query tree that exist in the
        index represented by the given ixreaderder.
        
        This method references the IndexReader to expand Prefix and Wildcard
        queries, and only adds terms that actually exist in the index (unless
        reverse=True).
        
        >>> ixreader = my_index.reader()
        >>> q = And([Or([Term("content", u"render"),
        ...             Term("content", u"rendering")]),
        ...             Prefix("path", u"/a/")])
        >>> q.existing_terms(ixreader, termset)
        set([("content", u"render"), ("path", u"/a/b"), ("path", u"/a/c")])
        
        :param ixreader: A :class:`whoosh.reading.IndexReader` object.
        :param reverse: If True, this method adds *missing* terms rather than
            *existing* terms to the set.
        :param phrases: Whether to add words found in Phrase queries.
        :rtype: set
        """

        if termset is None:
            termset = set()
        self._existing_terms(ixreader, termset, reverse=reverse,
                             phrases=phrases)
        return termset

    def field(self):
        """Returns the field this query matches in, or None if this query does
        not match in a single field.
        """
        
        return self.fieldname

    def estimate_size(self, ixreader):
        """Returns an estimate of how many documents this query could
        potentially match (for example, the estimated size of a simple term
        query is the document frequency of the term). It is permissible to
        overestimate, but not to underestimate.
        """
        raise NotImplementedError

    def estimate_min_size(self, ixreader):
        """Returns an estimate of the minimum number of documents this query
        could potentially match.
        """
        
        return self.estimate_size(ixreader)

    def matcher(self, searcher):
        """Returns a :class:`~whoosh.matching.Matcher` object you can use to
        retrieve documents and scores matching this query.
        
        :rtype: :class:`whoosh.matching.Matcher`
        """
        raise NotImplementedError

    def docs(self, searcher):
        """Returns an iterator of docnums matching this query.
        
        >>> searcher = my_index.searcher()
        >>> list(my_query.docs(searcher))
        [10, 34, 78, 103]
        
        :param searcher: A :class:`whoosh.searching.Searcher` object.
        """

        try:
            return self.matcher(searcher).all_ids()
        except TermNotFound:
            return iter([])

    def normalize(self):
        """Returns a recursively "normalized" form of this query. The
        normalized form removes redundancy and empty queries. This is called
        automatically on query trees created by the query parser, but you may
        want to call it yourself if you're writing your own parser or building
        your own queries.
        
        >>> q = And([And([Term("f", u"a"),
        ...               Term("f", u"b")]),
        ...               Term("f", u"c"), Or([])])
        >>> q.normalize()
        And([Term("f", u"a"), Term("f", u"b"), Term("f", u"c")])
        
        Note that this returns a *new, normalized* query. It *does not* modify
        the original query "in place".
        """
        return self

    def simplify(self, ixreader):
        """Returns a recursively simplified form of this query, where
        "second-order" queries (such as Prefix and Variations) are re-written
        into lower-level queries (such as Term and Or).
        """
        return self


class WrappingQuery(Query):
    def __init__(self, child):
        self.child = child
    
    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.child)
    
    def __hash__(self):
        return hash(self.__class__.__name__) ^ hash(self.child)
    
    def is_leaf(self):
        return False
    
    def apply(self, fn):
        return self.__class__(fn(self.child))
    
    def all_terms(self, termset=None, phrases=True):
        return self.child.all_terms(termset=termset, phrases=phrases)
    
    def existing_terms(self, ixreader, termset=None, reverse=False,
                       phrases=True):
        return self.child.existing_terms(ixreader, termset=termset,
                                         reverse=reverse, phrases=phrases)
    
    def field(self):
        return self.child.field()
    
    def estimate_size(self, ixreader):
        return self.child.estimate_size(ixreader)
    
    def estimate_min_size(self, ixreader):
        return self.child.estimate_min_size(ixreader)
    
    def matcher(self, searcher):
        return self.child.matcher(searcher)
    

class CompoundQuery(Query):
    """Abstract base class for queries that combine or manipulate the results
    of multiple sub-queries .
    """

    def __init__(self, subqueries, boost=1.0):
        self.subqueries = subqueries
        self.boost = boost

    def __repr__(self):
        r = "%s(%r" % (self.__class__.__name__, self.subqueries)
        if self.boost != 1:
            r += ", boost=%s" % self.boost
        r += ")"
        return r

    def __unicode__(self):
        r = u"("
        r += (self.JOINT).join([unicode(s) for s in self.subqueries])
        r += u")"
        return r

    def __eq__(self, other):
        return other and self.__class__ is other.__class__ and\
        self.subqueries == other.subqueries and\
        self.boost == other.boost

    def __getitem__(self, i):
        return self.subqueries.__getitem__(i)

    def __len__(self):
        return len(self.subqueries)

    def __iter__(self):
        return iter(self.subqueries)

    def __hash__(self):
        h = hash(self.__class__.__name__) ^ hash(self.boost)
        for q in self.subqueries:
            h ^= hash(q)
        return h

    def is_leaf(self):
        return False

    def apply(self, fn):
        return self.__class__([fn(q) for q in self.subqueries],
                              boost=self.boost)

    def field(self):
        if self.subqueries:
            f = self.subqueries[0].field()
            if all(q.field() == f for q in self.subqueries[1:]):
                return f

    def estimate_size(self, ixreader):
        return sum(q.estimate_size(ixreader) for q in self.subqueries)
    
    def estimate_min_size(self, ixreader):
        subs, nots = self._split_queries()
        subs_min = min(q.estimate_min_size(ixreader) for q in subs)
        if nots:
            nots_sum = sum(q.estimate_size(ixreader) for q in nots)
            subs_min = max(0, subs_min - nots_sum)
        return subs_min

    def _all_terms(self, termset, phrases=True):
        for q in self.subqueries:
            q.all_terms(termset, phrases=phrases)

    def _existing_terms(self, ixreader, termset, reverse=False, phrases=True):
        for q in self.subqueries:
            q.existing_terms(ixreader, termset, reverse=reverse,
                             phrases=phrases)

    def normalize(self):
        # Normalize subqueries and merge nested instances of this class
        subqueries = []
        for s in self.subqueries:
            s = s.normalize()
            if isinstance(s, self.__class__):
                subqueries += [ss.normalize() for ss in s.subqueries]
            else:
                subqueries.append(s)
        
        if all(q is NullQuery for q in subqueries):
            return NullQuery

        if any((isinstance(q, Every) and q.fieldname is None) for q in subqueries):
            return Every()

        # Merge ranges and Everys
        everyfields = set()
        i = 0
        while i < len(subqueries):
            q = subqueries[i]
            f = q.field()
            if f in everyfields:
                subqueries.pop(i)
                continue
            
            if isinstance(q, (TermRange, NumericRange)):
                j = i + 1
                while j < len(subqueries):
                    if q.overlaps(subqueries[j]):
                        qq = subqueries.pop(j)
                        q = q.merge(qq, intersect=self.intersect_merge)
                    else:
                        j += 1
                q = subqueries[i] = q.normalize()
            
            if isinstance(q, Every):
                everyfields.add(q.fieldname)
            i += 1

        # Eliminate duplicate queries
        subqs = []
        seenqs = set()
        for s in subqueries:
            if (not isinstance(s, Every) and s.field() in everyfields):
                continue
            if s in seenqs:
                continue
            seenqs.add(s)
            subqs.append(s)

        # Remove NullQuerys
        subqs = [q for q in subqs if q is not NullQuery]

        if not subqs:
            return NullQuery
        
        if len(subqs) == 1:
            sub = subqs[0]
            if not (self.boost == 1.0 and sub.boost == 1.0):
                sub = sub.copy()
                sub.boost *= self.boost
            return sub

        return self.__class__(subqs, boost=self.boost)

    def _split_queries(self):
        subs = [q for q in self.subqueries if not isinstance(q, Not)]
        nots = [q.query for q in self.subqueries if isinstance(q, Not)]
        return (subs, nots)

    def simplify(self, ixreader):
        subs, nots = self._split_queries()

        if subs:
            subs = self.__class__([subq.simplify(ixreader) for subq in subs],
                                  boost=self.boost).normalize()
            if nots:
                nots = Or(nots).simplify().normalize()
                return AndNot(subs, nots)
            else:
                return subs
        else:
            return NullQuery

    def _matcher(self, matchercls, searcher, **kwargs):
        # Pull any queries inside a Not() out into their own list
        subs, nots = self._split_queries()
        
        if not subs:
            return NullMatcher()
        
        # Sort the list of subqueries by their estimated size
        r = searcher.reader()
        subs.sort(key=lambda q: q.estimate_size(r))
        
        # Create a matcher from the list of subqueries
        subms = [subq.matcher(searcher) for subq in subs]
        if len(subms) == 1:
            m = subms[0]
        else:
            m = make_binary_tree(matchercls, subms)
        
        # If there were queries inside Not(), make a matcher for them and
        # wrap the matchers in an AndNotMatcher
        if nots:
            notms = [subq.matcher(searcher) for subq in nots]
            if len(notms) == 1:
                notm = notms[0]
            else:
                notm = make_binary_tree(UnionMatcher, notms)
                
            if notm.is_active():
                m = AndNotMatcher(m, notm)
        
        # If this query had a boost, add a wrapping matcher to apply the boost
        if self.boost != 1.0:
            m = WrappingMatcher(m, self.boost)
            
        return m


class MultiTerm(Query):
    """Abstract base class for queries that operate on multiple terms in the
    same field.
    """

    TOO_MANY_CLAUSES = 1024
    constantscore = False

    def _words(self, ixreader):
        raise NotImplementedError

    def simplify(self, ixreader):
        existing = [Term(self.fieldname, word, boost=self.boost)
                    for word in sorted(set(self._words(ixreader)))]
        if len(existing) == 1:
            return existing[0]
        elif existing:
            return Or(existing)
        else:
            return NullQuery

    def _all_terms(self, termset, phrases=True):
        pass

    def _existing_terms(self, ixreader, termset, reverse=False, phrases=True):
        fieldname = self.fieldname
        for word in self._words(ixreader):
            t = (fieldname, word)
            contains = t in ixreader
            if reverse:
                contains = not contains
            if contains:
                termset.add(t)

    def estimate_size(self, ixreader):
        return sum(ixreader.doc_frequency(self.fieldname, text)
                   for text in self._words(ixreader))

    def estimate_min_size(self, ixreader):
        return min(ixreader.doc_frequency(self.fieldname, text)
                   for text in self._words(ixreader))

    def matcher(self, searcher):
        fieldname = self.fieldname
        qs = [Term(fieldname, word) for word in self._words(searcher.reader())]
        if not qs:
            return NullMatcher()
        
        if len(qs) == 1:
            # If there's only one term, just use it
            q = qs[0]
        
        elif len(qs) > self.TOO_MANY_CLAUSES:
            # If there's so many clauses that an Or search would take forever,
            # trade memory for time and just put all the matching docs in a set
            # and serve it up as a ListMatcher
            docset = set()
            for q in qs:
                docset.update(q.matcher(searcher).all_ids())
            return ListMatcher(sorted(docset), all_weights=self.boost)
        
        else:
            # The default case: Or the terms together
            q = Or(qs)
        
        m = q.matcher(searcher)
        
        # Multiterm queries such as Prefix are usually just about finding
        # matching documents, not contributing to the score, so most of them
        # have constantscore set to True by default
        if self.constantscore:
            m = ConstantScoreMatcher(m, score=self.boost)
        
        return m
        

# Concrete classes

class Term(Query):
    """Matches documents containing the given term (fieldname+text pair).
    
    >>> Term("content", u"render")
    """

    __inittypes__ = dict(fieldname=str, text=unicode, boost=float)

    def __init__(self, fieldname, text, boost=1.0):
        self.fieldname = fieldname
        self.text = text
        self.boost = boost

    def __eq__(self, other):
        return (other
                and self.__class__ is other.__class__
                and
                self.fieldname == other.fieldname
                and self.text == other.text
                and self.boost == other.boost)

    def __repr__(self):
        r = "%s(%r, %r" % (self.__class__.__name__, self.fieldname, self.text)
        if self.boost != 1.0:
            r += ", boost=%s" % self.boost
        r += ")"
        return r

    def __unicode__(self):
        t = u"%s:%s" % (self.fieldname, self.text)
        if self.boost != 1:
            t += u"^" + unicode(self.boost)
        return t
    
    def __hash__(self):
        return hash(self.fieldname) ^ hash(self.text) ^ hash(self.boost)

    def _all_terms(self, termset, phrases=True):
        termset.add((self.fieldname, self.text))

    def _existing_terms(self, ixreader, termset, reverse=False, phrases=True):
        fieldname, text = self.fieldname, self.text
        contains = (fieldname, text) in ixreader
        if reverse:
            contains = not contains
        if contains:
            termset.add((fieldname, text))

    def replace(self, oldtext, newtext):
        q = self.copy()
        if q.text == oldtext:
            q.text = newtext
        return q

    def estimate_size(self, ixreader):
        return ixreader.doc_frequency(self.fieldname, self.text)

    def matcher(self, searcher):
        try:
            m = searcher.postings(self.fieldname, self.text)
            if self.boost != 1:
                m = WrappingMatcher(m, boost=self.boost)
                
            return m
        except TermNotFound:
            return NullMatcher()


class And(CompoundQuery):
    """Matches documents that match ALL of the subqueries.
    
    >>> And([Term("content", u"render"),
    ...      Term("content", u"shade"),
    ...      Not(Term("content", u"texture"))])
    >>> # You can also do this
    >>> Term("content", u"render") & Term("content", u"shade")
    """

    # This is used by the superclass's __unicode__ method.
    JOINT = " AND "
    intersect_merge = True

    def estimate_size(self, ixreader):
        return min(q.estimate_size(ixreader) for q in self.subqueries)

    def matcher(self, searcher):
        return self._matcher(IntersectionMatcher, searcher)


class Or(CompoundQuery):
    """Matches documents that match ANY of the subqueries.
    
    >>> Or([Term("content", u"render"),
    ...     And([Term("content", u"shade"), Term("content", u"texture")]),
    ...     Not(Term("content", u"network"))])
    >>> # You can also do this
    >>> Term("content", u"render") | Term("content", u"shade")
    """

    # This is used by the superclass's __unicode__ method.
    JOINT = " OR "
    intersect_merge = False

    def __init__(self, subqueries, boost=1.0, minmatch=0):
        CompoundQuery.__init__(self, subqueries, boost=boost)
        self.minmatch = minmatch

    def __unicode__(self):
        r = u"("
        r += (self.JOINT).join([unicode(s) for s in self.subqueries])
        r += u")"
        if self.minmatch:
            r += u">%s" % self.minmatch
        return r

    def normalize(self):
        norm = CompoundQuery.normalize(self)
        if norm.__class__ is self.__class__:
            norm.minmatch = self.minmatch
        return norm

    def matcher(self, searcher):
        return self._matcher(UnionMatcher, searcher)


class DisjunctionMax(CompoundQuery):
    """Matches all documents that match any of the subqueries, but scores each
    document using the maximum score from the subqueries.
    """

    def __init__(self, subqueries, boost=1.0, tiebreak=0.0):
        CompoundQuery.__init__(self, subqueries, boost=boost)
        self.tiebreak = tiebreak

    def __unicode__(self):
        r = u"DisMax("
        r += " ".join([unicode(s) for s in self.subqueries])
        r += u")"
        if self.tiebreak:
            s += u"~" + unicode(self.tiebreak)
        return r

    def normalize(self):
        norm = CompoundQuery.normalize(self)
        if norm.__class__ is self.__class__:
            norm.tiebreak = self.tiebreak
        return norm
    
    def matcher(self, searcher):
        return self._matcher(DisjunctionMaxMatcher, searcher,
                             tiebreak=self.tiebreak)


class Not(Query):
    """Excludes any documents that match the subquery.
    
    >>> # Match documents that contain 'render' but not 'texture'
    >>> And([Term("content", u"render"),
    ...      Not(Term("content", u"texture"))])
    >>> # You can also do this
    >>> Term("content", u"render") - Term("content", u"texture")
    """

    __inittypes__ = dict(query=Query)

    def __init__(self, query, boost=1.0):
        """
        :param query: A :class:`Query` object. The results of this query
            are *excluded* from the parent query.
        :param boost: Boost is meaningless for excluded documents but this
            keyword argument is accepted for the sake of a consistent interface.
        """

        self.query = query
        self.boost = boost

    def __eq__(self, other):
        return other and self.__class__ is other.__class__ and\
        self.query == other.query

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, repr(self.query))

    def __unicode__(self):
        return u"NOT " + unicode(self.query)

    def __hash__(self):
        return hash(self.__class__.__name__) ^ hash(self.query) ^ hash(self.boost)

    def is_leaf(self):
        return False

    def apply(self, fn):
        return self.__class__(fn(self.query))

    def normalize(self):
        query = self.query.normalize()
        if query is NullQuery:
            return NullQuery
        else:
            return self.__class__(query, boost=self.boost)

    def _all_terms(self, termset, phrases=True):
        self.query.all_terms(termset, phrases=phrases)

    def _existing_terms(self, ixreader, termset, reverse=False, phrases=True):
        self.query.existing_terms(ixreader, termset, reverse=reverse,
                                  phrases=phrases)

    def field(self):
        return None

    def estimate_size(self, ixreader):
        return ixreader.doc_count()
    
    def estimate_min_size(self, ixreader):
        return 1 if ixreader.doc_count() else 0

    def matcher(self, searcher):
        # Usually only called if Not is the root query. Otherwise, queries such
        # as And and Or do special handling of Not subqueries.
        reader = searcher.reader()
        child = self.query.matcher(searcher)
        return InverseMatcher(child, searcher.doc_count_all(),
                              missing=reader.is_deleted)


class PatternQuery(MultiTerm):
    """An intermediate base class for common methods of Prefix and Wildcard.
    """
    
    __inittypes__ = dict(fieldname=str, text=unicode, boost=float)

    def __init__(self, fieldname, text, boost=1.0, constantscore=True):
        self.fieldname = fieldname
        self.text = text
        self.boost = boost
        self.constantscore = constantscore
        
    def __eq__(self, other):
        return (other and self.__class__ is other.__class__
                and self.fieldname == other.fieldname
                and self.text == other.text and self.boost == other.boost
                and self.constantscore == other.constantscore)
        
    def __repr__(self):
        r = "%s(%r, %r" % (self.__class__.__name__, self.fieldname, self.text)
        if self.boost != 1:
            r += ", boost=" + self.boost
        r += ")"
        return r
    
    def __hash__(self):
        return (hash(self.fieldname) ^ hash(self.text) ^ hash(self.boost)
                ^ hash(self.constantscore))
    

class Prefix(PatternQuery):
    """Matches documents that contain any terms that start with the given text.
    
    >>> # Match documents containing words starting with 'comp'
    >>> Prefix("content", u"comp")
    """

    def __unicode__(self):
        return "%s:%s*" % (self.fieldname, self.text)

    def _words(self, ixreader):
        return ixreader.expand_prefix(self.fieldname, self.text)


class Wildcard(PatternQuery):
    """Matches documents that contain any terms that match a wildcard
    expression.
    
    >>> Wildcard("content", u"in*f?x")
    """

    def __unicode__(self):
        return "%s:%s" % (self.fieldname, self.text)

    def _words(self, ixreader):
        exp = re.compile(fnmatch.translate(self.text))

        # Get the "prefix" -- the substring before the first wildcard.
        qm = self.text.find("?")
        st = self.text.find("*")
        if qm < 0 and st < 0:
            prefix = ""
        elif qm < 0:
            prefix = self.text[:st]
        elif st < 0:
            prefix = self.text[:qm]
        else:
            prefix = self.text[:min(st, qm)]
        
        if prefix:
            candidates = ixreader.expand_prefix(self.fieldname, prefix)
        else:
            candidates = ixreader.lexicon(self.fieldname)

        for text in candidates:
            if exp.match(text):
                yield text

    def normalize(self):
        # If there are no wildcard characters in this "wildcard", turn it into
        # a simple Term
        text = self.text
        if text == "*":
            return Every(self.fieldname, boost=self.boost)
        if "*" not in text and "?" not in text:
            # If no wildcard chars, convert to a normal term.
            return Term(self.fieldname, self.text, boost=self.boost)
        elif ("?" not in text
              and text.endswith("*")
              and text.find("*") == len(text) - 1
              and (len(text) < 2 or text[-2] != "\\")):
            # If the only wildcard char is an asterisk at the end, convert to a
            # Prefix query.
            return Prefix(self.fieldname, self.text[:-1], boost=self.boost)
        else:
            return self


class FuzzyTerm(MultiTerm):
    """Matches documents containing words similar to the given term.
    """

    __inittypes__ = dict(fieldname=str, text=unicode, boost=float,
                         minsimilarity=float, prefixlength=int)

    def __init__(self, fieldname, text, boost=1.0, minsimilarity=0.5,
                 prefixlength=1, constantscore=True):
        """
        :param fieldname: The name of the field to search.
        :param text: The text to search for.
        :param boost: A boost factor to apply to scores of documents matching
            this query.
        :param minsimilarity: The minimum similarity ratio to match. 1.0 is the
            maximum (an exact match to 'text').
        :param prefixlength: The matched terms must share this many initial
            characters with 'text'. For example, if text is "light" and
            prefixlength is 2, then only terms starting with "li" are checked
            for similarity.
        """

        self.fieldname = fieldname
        self.text = text
        self.boost = boost
        self.minsimilarity = minsimilarity
        self.prefixlength = prefixlength
        self.constantscore = constantscore

    def __eq__(self, other):
        return (other and self.__class__ is other.__class__
                and self.fieldname == other.fieldname
                and self.text == other.text
                and self.minsimilarity == other.minsimilarity
                and self.prefixlength == other.prefixlength
                and self.boost == other.boost
                and self.constantscore == other.constantscore)

    def __repr__(self):
        r = "%s(%r, %r, boost=%f, minsimilarity=%f, prefixlength=%d)"
        return r % (self.__class__.__name__, self.fieldname, self.text,
                    self.boost, self.minsimilarity, self.prefixlength)

    def __unicode__(self):
        r = u"~" + self.text
        if self.boost != 1.0:
            r += "^%f" % self.boost
        return r

    def __hash__(self):
        return (hash(self.fieldname) ^ hash(self.text) ^ hash(self.boost)
                ^ hash(self.minsimilarity) ^ hash(self.prefixlength)
                ^ hash(self.constantscore))

    def _all_terms(self, termset, phrases=True):
        termset.add((self.fieldname, self.text))

    def _words(self, ixreader):
        text = self.text
        minsim = self.minsimilarity
        for term in ixreader.expand_prefix(self.fieldname,
                                           text[:self.prefixlength]):
            if text == term:
                yield term
            elif relative(text, term) > minsim:
                yield term


class RangeMixin(object):
    # Contains methods shared by TermRange and NumericRange
    
    def __repr__(self):
        return ('%s(%r, %r, %r, %s, %s, boost=%s, constantscore=%s)'
                % (self.__class__.__name__, self.fieldname, self.start,
                   self.end, self.startexcl, self.endexcl, self.boost,
                   self.constantscore))
    
    def __unicode__(self):
        startchar = "{" if self.startexcl else "["
        endchar = "}" if self.endexcl else "]"
        start = '' if self.start is None else self.start
        end = '' if self.end is None else self.end
        return u"%s:%s%s TO %s%s" % (self.fieldname, startchar, start, end,
                                     endchar)
    
    def __eq__(self, other):
        return (other and self.__class__ is other.__class__
                and self.fieldname == other.fieldname
                and self.start == other.start
                and self.end == other.end
                and self.startexcl == other.startexcl
                and self.endexcl == other.endexcl
                and self.boost == other.boost)
    
    def __hash__(self):
        return (hash(self.fieldname) ^ hash(self.start) ^ hash(self.startexcl)
                ^ hash(self.end) ^ hash(self.endexcl) ^ hash(self.boost))
    
    def _comparable_start(self):
        if self.start is None:
            return (Lowest, 0)
        else:
            second = 1 if self.startexcl else 0
            return (self.start, second)
        
    def _comparable_end(self):
        if self.end is None:
            return (Highest, 0)
        else:
            second = -1 if self.endexcl else 0
            return (self.end, second)

    def overlaps(self, other):
        if not isinstance(other, TermRange):
            return False
        if self.fieldname != other.fieldname:
            return False
        
        start1 = self._comparable_start()
        start2 = other._comparable_start()
        end1 = self._comparable_end()
        end2 = other._comparable_end()
        
        return ((start1 >= start2 and start1 <= end2)
                or (end1 >= start2 and end1 <= end2)
                or (start2 >= start1 and start2 <= end1)
                or (end2 >= start1 and end2 <= end1))

    def merge(self, other, intersect=True):
        assert self.fieldname == other.fieldname
        
        start1 = self._comparable_start()
        start2 = other._comparable_start()
        end1 = self._comparable_end()
        end2 = other._comparable_end()
        
        if start1 >= start2 and end1 <= end2:
            start = start2
            end = end2
        elif start2 >= start1 and end2 <= end1:
            start = start1
            end = end1
        elif intersect:
            start = max(start1, start2)
            end = min(end1, end2)
        else:
            start = min(start1, start2)
            end = max(end1, end2)
        
        startval = None if start[0] is Lowest else start[0]
        startexcl = start[1] == 1
        endval = None if end[0] is Highest else end[0]
        endexcl = end[1] == -1
        
        boost = max(self.boost, other.boost)
        constantscore = self.constantscore or other.constantscore
        
        return self.__class__(self.fieldname, startval, endval, startexcl,
                              endexcl, boost=boost, constantscore=constantscore)


class TermRange(RangeMixin, MultiTerm):
    """Matches documents containing any terms in a given range.
    
    >>> # Match documents where the indexed "id" field is greater than or equal
    >>> # to 'apple' and less than or equal to 'pear'.
    >>> TermRange("id", u"apple", u"pear")
    """

    def __init__(self, fieldname, start, end, startexcl=False, endexcl=False,
                 boost=1.0, constantscore=True):
        """
        :param fieldname: The name of the field to search.
        :param start: Match terms equal to or greater than this.
        :param end: Match terms equal to or less than this.
        :param startexcl: If True, the range start is exclusive. If False, the
            range start is inclusive.
        :param endexcl: If True, the range end is exclusive. If False, the
            range end is inclusive.
        :param boost: Boost factor that should be applied to the raw score of
            results matched by this query.
        """

        self.fieldname = fieldname
        self.start = start
        self.end = end
        self.startexcl = startexcl
        self.endexcl = endexcl
        self.boost = boost
        self.constantscore = constantscore

    def normalize(self):
        if self.start in ('', None) and self.end in (u'\uffff', None):
            return Every(self.fieldname, boost=self.boost)
        elif self.start == self.end:
            if self.startexcl or self.endexcl:
                return NullQuery
            return Term(self.fieldname, self.start, boost=self.boost)
        else:
            return TermRange(self.fieldname, self.start, self.end,
                             self.startexcl, self.endexcl,
                             boost=self.boost)

    def replace(self, oldtext, newtext):
        q = self.copy()
        if q.start == oldtext:
            q.start = newtext
        if q.end == oldtext:
            q.end = newtext
        return q
    
    def _words(self, ixreader):
        fieldname = self.fieldname
        start = '' if self.start is None else self.start
        end = u'\uFFFF' if self.end is None else self.end
        startexcl = self.startexcl
        endexcl = self.endexcl

        for fname, t, _, _ in ixreader.iter_from(fieldname, start):
            if fname != fieldname:
                break
            if t == start and startexcl:
                continue
            if t == end and endexcl:
                break
            if t > end:
                break
            yield t


class NumericRange(RangeMixin, Query):
    """A range query for NUMERIC fields. Takes advantage of tiered indexing
    to speed up large ranges by matching at a high resolution at the edges of
    the range and a low resolution in the middle.
    
    >>> # Match numbers from 10 to 5925 in the "number" field.
    >>> nr = NumericRange("number", 10, 5925)
    """
    
    def __init__(self, fieldname, start, end, startexcl=False, endexcl=False,
                 boost=1.0, constantscore=True):
        """
        :param fieldname: The name of the field to search.
        :param start: Match terms equal to or greater than this number. This
            should be a number type, not a string.
        :param end: Match terms equal to or less than this number. This should
            be a number type, not a string.
        :param startexcl: If True, the range start is exclusive. If False, the
            range start is inclusive.
        :param endexcl: If True, the range end is exclusive. If False, the
            range end is inclusive.
        :param boost: Boost factor that should be applied to the raw score of
            results matched by this query.
        :param constantscore: If True, the compiled query returns a constant
            score (the value of the ``boost`` keyword argument) instead of
            actually scoring the matched terms. This gives a nice speed boost
            and won't affect the results in most cases since numeric ranges
            will almost always be used as a filter.
        """

        self.fieldname = fieldname
        self.start = start
        self.end = end
        self.startexcl = startexcl
        self.endexcl = endexcl
        self.boost = boost
        self.constantscore = constantscore
    
    def simplify(self, ixreader):
        return self._compile_query(ixreader).simplify(ixreader)
    
    def estimate_size(self, ixreader):
        return self._compile_query(ixreader).estimate_size(ixreader)
    
    def estimate_min_size(self, ixreader):
        return self._compile_query(ixreader).estimate_min_size(ixreader)
    
    def docs(self, searcher):
        q = self._compile_query(searcher.reader())
        return q.docs(searcher)
    
    def _compile_query(self, ixreader):
        from whoosh.fields import NUMERIC
        from whoosh.support.numeric import tiered_ranges
        
        field = ixreader.schema[self.fieldname]
        if not isinstance(field, NUMERIC):
            raise Exception("NumericRange: field %r is not numeric" % self.fieldname)
        
        start = field.prepare_number(self.start)
        end = field.prepare_number(self.end)
        
        subqueries = []
        # Get the term ranges for the different resolutions
        for starttext, endtext in tiered_ranges(field.type, field.signed,
                                                start, end, field.shift_step,
                                                self.startexcl, self.endexcl):
            if starttext == endtext:
                subq = Term(self.fieldname, starttext)
            else:
                subq = TermRange(self.fieldname, starttext, endtext)
            subqueries.append(subq)
        
        if len(subqueries) == 1:
            q = subqueries[0] 
        elif subqueries:
            q = Or(subqueries, boost=self.boost)
        else:
            return NullQuery
        
        if self.constantscore:
            q = ConstantScoreQuery(q, self.boost)
        return q
        
    def matcher(self, searcher):
        q = self._compile_query(searcher.reader())
        return q.matcher(searcher)


class DateRange(NumericRange):
    """This is a very thin subclass of :class:`NumericRange` that only
    overrides the initializer and ``__repr__()`` methods to work with datetime
    objects instead of numbers. Internally this object converts the datetime
    objects it's created with to numbers and otherwise acts like a
    ``NumericRange`` query.
    
    >>> DateRange("date", datetime(2010, 11, 3, 3, 0), datetime(2010, 11, 3, 17, 59))
    """
    
    def __init__(self, fieldname, start, end, startexcl=False, endexcl=False,
                 boost=1.0, constantscore=True):
        self.startdate = start
        self.enddate = end
        if start:
            start = datetime_to_long(start)
        if end:
            end = datetime_to_long(end)
        super(DateRange, self).__init__(fieldname, start, end,
                                        startexcl=startexcl, endexcl=endexcl,
                                        boost=boost, constantscore=constantscore)
    
    def __repr__(self):
        return '%s(%r, %r, %r, %s, %s, boost=%s)' % (self.__class__.__name__,
                                           self.fieldname,
                                           self.startdate, self.enddate,
                                           self.startexcl, self.endexcl,
                                           self.boost)
    

class Variations(MultiTerm):
    """Query that automatically searches for morphological variations of the
    given word in the same field.
    """

    def __init__(self, fieldname, text, boost=1.0):
        self.fieldname = fieldname
        self.text = text
        self.boost = boost

    def __repr__(self):
        r = "%s(%r, %r" % (self.__class__.__name__, self.fieldname, self.text)
        if self.boost != 1:
            r += ", boost=%s" % self.boost
        r += ")"
        return r

    def __eq__(self, other):
        return (other and self.__class__ is other.__class__
                and self.fieldname == other.fieldname
                and self.text == other.text and self.boost == other.boost)

    def __hash__(self):
        return hash(self.fieldname) ^ hash(self.text) ^ hash(self.boost)

    def _all_terms(self, termset, phrases=True):
        termset.add(self.text)

    def _existing_terms(self, ixreader, termset, reverse=False, phrases=True):
        for word in variations(self.text):
            t = (self.fieldname, word)
            contains = t in ixreader
            if reverse:
                contains = not contains
            if contains:
                termset.add(t)

    def _words(self, ixreader):
        fieldname = self.fieldname
        return [word for word in variations(self.text)
                if (fieldname, word) in ixreader]

    def __unicode__(self):
        return u"%s:<%s>" % (self.fieldname, self.text)

    def replace(self, oldtext, newtext):
        q = self.copy()
        if q.text == oldtext:
            q.text = newtext
        return q


class Phrase(Query):
    """Matches documents containing a given phrase."""

    def __init__(self, fieldname, words, slop=1, boost=1.0):
        """
        :param fieldname: the field to search.
        :param words: a list of words (unicode strings) in the phrase.
        :param slop: the number of words allowed between each "word" in the
            phrase; the default of 1 means the phrase must match exactly.
        :param boost: a boost factor that to apply to the raw score of
            documents matched by this query.
        """

        self.fieldname = fieldname
        self.words = words
        self.slop = slop
        self.boost = boost

    def __eq__(self, other):
        return (other and self.__class__ is other.__class__ and
                self.fieldname == other.fieldname and self.words == other.words
                and self.slop == other.slop and self.boost == other.boost)
        
    def __repr__(self):
        return "%s(%r, %r, slop=%s, boost=%f)" % (self.__class__.__name__,
                                                  self.fieldname, self.words,
                                                  self.slop, self.boost)

    def __unicode__(self):
        return u'%s:"%s"' % (self.fieldname, u" ".join(self.words))

    def __hash__(self):
        h = hash(self.fieldname) ^ hash(self.slop) ^ hash(self.boost)
        for w in self.words:
            h ^= hash(w)
        return h

    def copy(self):
        # Need to override the default shallow copy here to do a copy of the
        # self.words list
        return self.__class__(self.fieldname, self.words[:], boost=self.boost)

    def _all_terms(self, termset, phrases=True):
        if phrases:
            fieldname = self.fieldname
            for word in self.words:
                termset.add((fieldname, word))

    def _existing_terms(self, ixreader, termset, reverse=False, phrases=True):
        if phrases:
            fieldname = self.fieldname
            for word in self.words:
                contains = (fieldname, word) in ixreader
                if reverse:
                    contains = not contains
                if contains:
                    termset.add((fieldname, word))

    def normalize(self):
        if not self.words:
            return NullQuery
        if len(self.words) == 1:
            return Term(self.fieldname, self.words[0])

        words = [w for w in self.words if w is not None]
        return self.__class__(self.fieldname, words, slop=self.slop,
                              boost=self.boost)

    def replace(self, oldtext, newtext):
        q = self.copy()
        for i in xrange(len(q.words)):
            if q.words[i] == oldtext:
                q.words[i] = newtext
        return q

    def _and_query(self):
        fn = self.fieldname
        return And([Term(fn, word) for word in self.words])

    def estimate_size(self, ixreader):
        return self._and_query().estimate_size(ixreader)

    def estimate_min_size(self, ixreader):
        return self._and_query().estimate_min_size(ixreader)

    def matcher(self, searcher):
        fieldname = self.fieldname
        reader = searcher.reader()

        # Shortcut the query if one of the words doesn't exist.
        for word in self.words:
            if (fieldname, word) not in reader:
                return NullMatcher()
        
        field = searcher.schema[fieldname]
        if not field.format or not field.format.supports("positions"):
            raise QueryError("Phrase search: %r field has no positions"
                             % self.fieldname)
        
        # Construct a tree of SpanNear queries representing the words in the
        # phrase and return its matcher
        from whoosh.spans import SpanNear
        q = SpanNear.phrase(fieldname, self.words, slop=self.slop)
        m = q.matcher(searcher)
        if self.boost != 1.0:
            m = WrappingMatcher(m, boost=self.boost)
        return m


class Ordered(And):
    """Matches documents containing a list of sub-queries in the given order.
    """

    JOINT = " BEFORE "
    
    def matcher(self, searcher):
        from spans import SpanBefore
        return self._matcher(SpanBefore._Matcher, searcher)


class Every(Query):
    """A query that matches every document containing any term in a given
    field. If you don't specify a field, the query matches every document.
    
    >>> # Match any documents with something in the "path" field
    >>> q = Every("path")
    >>> # Matcher every document
    >>> q = Every()
    
    The unfielded form (matching every document) is efficient.
    
    The fielded is more efficient than a prefix query with an empty prefix or a
    '*' wildcard, but it can still be very slow on large indexes. It requires
    the searcher to read the full posting list of every term in the given
    field.
    
    Instead of using this query it is much more efficient when you create the
    index to include a single term that appears in all documents that have the
    field you want to match.
    
    For example, instead of this::
    
        # Match all documents that have something in the "path" field
        q = Every("path")
        
    Do this when indexing::
    
        # Add an extra field that indicates whether a document has a path
        schema = fields.Schema(path=fields.ID, has_path=fields.ID)
        
        # When indexing, set the "has_path" field based on whether the document
        # has anything in the "path" field
        writer.add_document(text=text_value1)
        writer.add_document(text=text_value2, path=path_value2, has_path="t")
    
    Then to find all documents with a path::
    
        q = Term("has_path", "t")
    """

    def __init__(self, fieldname=None, boost=1.0):
        """
        :param fieldname: the name of the field to match, or ``None`` or ``*``
            to match all documents.
        """
        
        if not fieldname or fieldname == "*":
            fieldname = None
        self.fieldname = fieldname
        self.boost = boost

    def __repr__(self):
        return "%s(%r, boost=%s)" % (self.__class__.__name__, self.fieldname,
                                     self.boost)

    def __eq__(self, other):
        return (other and self.__class__ is other.__class__
                and self.fieldname == other.fieldname
                and self.boost == other.boost)

    def __unicode__(self):
        return u"%s:*" % self.fieldname

    def __hash__(self):
        return hash(self.fieldname)

    def estimate_size(self, ixreader):
        return ixreader.doc_count()

    def matcher(self, searcher):
        fieldname = self.fieldname
        reader = searcher.reader()
        
        if fieldname in (None, "", "*"):
            doclist = list(searcher.reader().all_doc_ids())
        elif reader.supports_caches() and reader.fieldcache_available(self.fieldname):
            # If the reader has a field cache, use it to quickly get the list
            # of documents that have a value for this field
            fc = reader.fieldcache(self.fieldname)
            doclist = [docnum for docnum, ord in fc.ords() if ord != 0]
        else:
            # This is a hacky hack, but just create an in-memory set of all the
            # document numbers of every term in the field. This is SLOOOW for
            # large indexes
            doclist = set()
            for text in searcher.lexicon(fieldname):
                pr = searcher.postings(fieldname, text)
                doclist.update(pr.all_ids())
            doclist = sorted(doclist)
        
        return ListMatcher(doclist, all_weights=self.boost)

            
class NullQuery(Query):
    "Represents a query that won't match anything."
    def __call__(self):
        return self
    
    def __repr__(self):
        return "<%s>" % (self.__class__.__name__, )
    
    def __hash__(self):
        return id(self)
    
    def copy(self):
        return self
    
    def field(self):
        return None
    
    def estimate_size(self, ixreader):
        return 0
    
    def normalize(self):
        return self
    
    def simplify(self, ixreader):
        return self
    
    def docs(self, searcher):
        return []
    
    def matcher(self, searcher):
        return NullMatcher()


NullQuery = NullQuery()


class ConstantScoreQuery(WrappingQuery):
    """Wraps a query and uses a matcher that always gives a constant score
    to all matching documents. This is a useful optimization when you don't
    care about scores from a certain branch of the query tree because it is
    simply acting as a filter. See also the :class:`AndMaybe` query.
    """
    
    def __init__(self, child, score=1.0):
        super(ConstantScoreQuery, self).__init__(child)
        self.score = score
    
    def __eq__(self, other):
        return (other and self.__class__ is other.__class__
                and self.child == other.child and self.score == other.score)
    
    def __hash__(self):
        return hash(self.child) ^ hash(self.score)
    
    def apply(self, fn):
        return self.__class__(fn(self.child), self.score)
    
    def matcher(self, searcher):
        m = self.child.matcher(searcher)
        if isinstance(m, NullMatcher):
            return m
        else:
            ids = array("I", m.all_ids())
            return ListMatcher(ids, all_weights=self.score)
    

class WeightingQuery(WrappingQuery):
    """Wraps a query and specifies a custom weighting model to apply to the
    wrapped branch of the query tree. This is useful when you want to score
    parts of the query using criteria that don't apply to the rest of the
    query.
    """
    
    def __init__(self, child, model, fieldname=None, text=None):
        super(WeightingQuery, self).__init__(child)
        self.model = model
        self.fieldname = fieldname
        self.text = text
    
    def __eq__(self, other):
        return (other and self.__class__ is other.__class__
                and self.child == other.child
                and self.model == other.model
                and self.fieldname == other.fieldname
                and self.text == other.text)
    
    def __hash__(self):
        return hash(self.child) ^ hash(self.fieldname) ^ hash(self.text)
    
    def apply(self, fn):
        return self.__class__(fn(self.child), self.model, self.fieldname,
                              self.text)
    
    def matcher(self, searcher):
        m = self.child.matcher(searcher)
        scorer = self.model.scorer(searcher, self.fieldname, self.text)
        if isinstance(m, NullMatcher):
            return m
        else:
            return WeightingQuery.CustomScorerMatcher(m, scorer)
    
    class CustomScorerMatcher(WrappingMatcher):
        def __init__(self, child, scorer):
            super(WeightingQuery, self).__init__(child)
            self.scorer = scorer
        
        def copy(self):
            return self.__class__(self.child.copy(), self.scorer)
        
        def _replacement(self, newchild):
            return self.__class__(newchild, self.scorer)
        
        def supports_quality(self):
            return self.scorer.supports_quality()
        
        def quality(self):
            return self.scorer.quality(self)
        
        def block_quality(self):
            return self.scorer.block_quality(self)
        
        def score(self):
            return self.scorer.score(self)


class BinaryQuery(CompoundQuery):
    """Base class for binary queries (queries which are composed of two
    sub-queries). Subclasses should set the ``matcherclass`` attribute or
    override ``matcher()``, and may also need to override ``normalize()``,
    ``estimate_size()``, and/or ``estimate_min_size()``.
    """
    
    def __init__(self, a, b, boost=1.0):
        self.a = a
        self.b = b
        self.subqueries = (a, b)
        self.boost = boost

    def __eq__(self, other):
        return (other and self.__class__ is other.__class__
                and self.a == other.a and self.b == other.b
                and self.boost == other.boost)
    
    def __hash__(self):
        return (hash(self.__class__.__name__) ^ hash(self.a) ^ hash(self.b)
                ^ hash(self.boost))
    
    def apply(self, fn):
        return self.__class__(fn(self.a), fn(self.b), boost=self.boost)
    
    def field(self):
        f = self.a.field()
        if self.b.field() == f:
            return f
    
    def normalize(self):
        a = self.a.normalize()
        b = self.b.normalize()
        if a is NullQuery and b is NullQuery:
            return NullQuery
        elif a is NullQuery:
            return b
        elif b is NullQuery:
            return a
    
        return self.__class__(a, b, boost=self.boost)
    
    def matcher(self, searcher):
        return self.matcherclass(self.a.matcher(searcher),
                                 self.b.matcher(searcher))


class Require(BinaryQuery):
    """Binary query returns results from the first query that also appear in
    the second query, but only uses the scores from the first query. This lets
    you filter results without affecting scores.
    """

    JOINT = " REQUIRE "
    matcherclass = RequireMatcher

    def estimate_size(self, ixreader):
        return self.b.estimate_size(ixreader)
    
    def estimate_min_size(self, ixreader):
        return self.b.estimate_min_size(ixreader)

    def normalize(self):
        a = self.a.normalize()
        b = self.b.normalize()
        if a is NullQuery or b is NullQuery:
            return NullQuery
        return self.__class__(a, b, boost=self.boost)
    
    def docs(self, searcher):
        return And(self.subqueries).docs(searcher)
    

class AndMaybe(BinaryQuery):
    """Binary query takes results from the first query. If and only if the
    same document also appears in the results from the second query, the score
    from the second query will be added to the score from the first query.
    """

    JOINT = " ANDMAYBE "
    matcherclass = AndMaybeMatcher

    def normalize(self):
        a = self.a.normalize()
        b = self.b.normalize()
        if a is NullQuery:
            return NullQuery
        if b is NullQuery:
            return a
        return self.__class__(a, b, boost=self.boost)

    def estimate_min_size(self, ixreader):
        return self.subqueries[0].estimate_min_size(ixreader)

    def docs(self, searcher):
        return self.subqueries[0].docs(searcher)


class AndNot(BinaryQuery):
    """Binary boolean query of the form 'a ANDNOT b', where documents that
    match b are removed from the matches for a.
    """

    JOINT = " ANDNOT "
    matcherclass = AndNotMatcher

    def normalize(self):
        a = self.a.normalize()
        b = self.b.normalize()

        if a is NullQuery:
            return NullQuery
        elif b is NullQuery:
            return a

        return self.__class__(a, b, boost=self.boost)

    def _all_terms(self, termset, phrases=True):
        self.a.all_terms(termset, phrases=phrases)

    def _existing_terms(self, ixreader, termset, reverse=False, phrases=True):
        self.a.existing_terms(ixreader, termset, reverse=reverse,
                              phrases=phrases)


class Otherwise(BinaryQuery):
    """A binary query that only matches the second clause if the first clause
    doesn't match any documents.
    """
    
    JOINT = " OTHERWISE "
    
    def matcher(self, searcher):
        m = self.a.matcher(searcher)
        if not m.is_active():
            m = self.b.matcher(searcher)
        return m


def BooleanQuery(required, should, prohibited):
    return AndNot(AndMaybe(And(required), Or(should)), Or(prohibited)).normalize()














import sys, os.path
from ConfigParser import ConfigParser
from optparse import OptionParser
from os import system

# Script to build and upload a release of Whoosh to PyPI and build
# and upload the 

def build_docs():
    system("python setup.py build_sphinx")

def upload_docs(user, server, base, version, build=True, latest=True):
    opts = {"user": user,
            "srv": server,
            "base": base,
            "ver": version}

    system('ssh %(user)s@%(srv)s "mkdir %(base)s/%(ver)s"' % opts)
    system("scp -r docs/build/html/* %(user)s@%(srv)s:%(base)s/%(ver)s" % opts)
    system('ssh %(user)s@%(srv)s "cd %(base)s;ln -s %(ver)s latest"' % opts)


def upload_pypi(tag=None):
    system("python setup.py sdist bdist_egg upload")
    if tag:
        tag = str(tag)
        opts = {"base": "http://svn.whoosh.ca/projects/whoosh",
                "tag": tag,
                "msg": "Tagging trunk as %s" % tag}
        
        system('svn copy %(base)s/trunk %(base)s/tags/%(tag)s -m "%(msg)s"' % opts)


if __name__ == '__main__':
    sys.path.insert(0, os.path.abspath("src"))
    from whoosh import __version__

    version = ".".join(str(n) for n in __version__)

    parser = OptionParser()
    parser.add_option("-c", "--config", dest="configfile",
                      help="Configuration file",
                      metavar="INIFILE",
                      default="whoosh.ini")
    
    parser.add_option("-d", "--no-docs", dest="dodocs",
                      help="Don't build or upload docs",
                      action="store_false",
                      default=True)
    
    parser.add_option("-D", "--no-build-docs", dest="builddocs",
                      help="Skip building docs",
                      action="store_false",
                      default=True)
    
    parser.add_option("-t", "--tag", dest="tag",
                      help="Tag the trunk as this",
                      default=None)
    
    (options, args) = parser.parse_args()
    
    cp = ConfigParser()
    cp.read(options.configfile)
    
    if options.dodocs:
        upload_docs(cp.get("website", "username"),
                    cp.get("website", "server"),
                    cp.get("website", "docbase"), version,
                    build=options.builddocs)

    upload_pypi(tag=options.tag)



import os
import unittest

from whoosh import fields
from whoosh.filedb import pools, filestore


class FakeTermIndex(object):
    def __init__(self):
        self.d = {}
    
    def add(self, term, info):
        fieldname, text = term
        weight, offset, postcount = info
        self.d[term] = info
    
    def __getattr__(self, name):
        raise Exception("FTI name=%s" % name)
    

class FakePostWriter(object):
    def __init__(self):
        self.l = []
        self.blockcount = 0
        self.blockids = self.blockweights = self.blockvalues = []
    
    def start(self, format):
        self.ps = []
        return len(self.l)
        
    def write(self, *args):
        self.ps.append(args)
    
    def finish(self):
        return len(self.ps)
    
    def cancel(self):
        pass
    
    def as_inline(self):
        return ([], [], [], None, None)
    
    @property
    def posttotal(self):
        return len(self.ps)
    
    def __getattr__(self, name):
        raise Exception("FPW name=%s" % name)


#class TestPool(unittest.TestCase):
#    def test_addpostings(self):
#        s = fields.Schema(text=fields.TEXT)
#        st = filestore.RamStorage()
#        
#        p = pools.TempfilePool(s)
#        p.add_posting("text", u"alfa", 0, 1.0, "\x00\x00\x00\x01")
#        p.add_posting("text", u"bravo", 0, 2.0, "\x00\x00\x00\x02")
#        p.add_posting("text", u"charlie", 0, 3.0, "\x00\x00\x00\x03")
#        p.add_field_length(0, "text", 6)
#        p.add_posting("text", u"bravo", 1, 4.0, "\x00\x00\x00\x04")
#        p.add_posting("text", u"charlie", 1, 5.0, "\x00\x00\x00\x05")
#        p.add_posting("text", u"delta", 1, 6.0, "\x00\x00\x00\x06")
#        p.add_field_length(1, "text", 15)
#        
#        p.dump_run()
#        
#        doccount = 2
#        lengthfile = st.create_file("test.len")
#        termtable = FakeTermIndex()
#        postwriter = FakePostWriter()
#        
#        p.finish(doccount, lengthfile, termtable, postwriter)



if __name__ == '__main__':
    unittest.main()




import unittest

from whoosh import fields, qparser
from whoosh.analysis import *
from whoosh.filedb.filestore import RamStorage


class TestAnalysis(unittest.TestCase):
    def test_regextokenizer(self):
        value = u"AAAaaaBBBbbbCCCcccDDDddd"
        
        rex = RegexTokenizer("[A-Z]+")
        self.assertEqual([t.text for t in rex(value)],
                         [u"AAA", u"BBB", u"CCC", u"DDD"])
        
        rex = RegexTokenizer("[A-Z]+", gaps=True)
        self.assertEqual([t.text for t in rex(value)],
                         [u"aaa", u"bbb", u"ccc", u"ddd"])
    
    def test_composition1(self):
        ca = RegexTokenizer() | LowercaseFilter()
        self.assertEqual(ca.__class__.__name__, "CompositeAnalyzer")
        self.assertEqual(ca[0].__class__.__name__, "RegexTokenizer")
        self.assertEqual(ca[1].__class__.__name__, "LowercaseFilter")
        self.assertEqual([t.text for t in ca(u"ABC 123")], ["abc", "123"])
    
    def test_composition2(self):
        ca = RegexTokenizer() | LowercaseFilter()
        sa = ca | StopFilter()
        self.assertEqual(len(sa), 3)
        self.assertEqual(sa.__class__.__name__, "CompositeAnalyzer")
        self.assertEqual(sa[0].__class__.__name__, "RegexTokenizer")
        self.assertEqual(sa[1].__class__.__name__, "LowercaseFilter")
        self.assertEqual(sa[2].__class__.__name__, "StopFilter")
        self.assertEqual([t.text for t in sa(u"The ABC 123")], ["abc", "123"])
    
    def test_composition3(self):
        sa = RegexTokenizer() | StopFilter()
        self.assertEqual(sa.__class__.__name__, "CompositeAnalyzer")
    
    def test_composing_functions(self):
        def filter(tokens):
            for t in tokens:
                t.text = t.text.upper()
                yield t
                
        analyzer = RegexTokenizer() | filter
        self.assertEqual([t.text for t in analyzer(u"abc def")], ["ABC", "DEF"])
    
    def test_shared_composition(self):
        shared = RegexTokenizer(r"\S+") | LowercaseFilter()
        
        ana1 = shared | NgramFilter(3)
        ana2 = shared | DoubleMetaphoneFilter()
        
        self.assertEqual([t.text for t in ana1(u"hello")], ["hel", "ell", "llo"])
        self.assertEqual([t.text for t in ana2(u"hello")], ["HL"])
    
    def test_multifilter(self):
        f1 = LowercaseFilter()
        f2 = PassFilter()
        mf = MultiFilter(a=f1, b=f2)
        ana = RegexTokenizer(r"\S+") | mf
        text = u"ALFA BRAVO CHARLIE"
        self.assertEqual([t.text for t in ana(text, mode="a")], ["alfa", "bravo", "charlie"])
        self.assertEqual([t.text for t in ana(text, mode="b")], ["ALFA", "BRAVO", "CHARLIE"])
    
    def test_intraword(self):
        iwf = IntraWordFilter(mergewords=True, mergenums=True)
        ana = RegexTokenizer(r"\S+") | iwf
        
        def check(text, ls):
            self.assertEqual([(t.pos, t.text) for t in ana(text)], ls)
            
        check(u"PowerShot", [(0, "Power"), (1, "Shot"), (1, "PowerShot")])
        check(u"A's+B's&C's", [(0, "A"), (1, "B"), (2, "C"), (2, "ABC")])
        check(u"Super-Duper-XL500-42-AutoCoder!", [(0, "Super"), (1, "Duper"), (2, "XL"),
                                                   (2, "SuperDuperXL"), (3, "500"), (4, "42"),
                                                   (4, "50042"), (5, "Auto"), (6, "Coder"),
                                                   (6, "AutoCoder")])
    
    def test_biword(self):
        ana = RegexTokenizer(r"\w+") | BiWordFilter()
        result = [t.copy() for t in ana(u"the sign of four",
                                        chars=True, positions=True)]
        self.assertEqual(["the-sign", "sign-of", "of-four"],
                         [t.text for t in result])
        self.assertEqual([(0, 8), (4, 11), (9, 16)],
                         [(t.startchar, t.endchar) for t in result])
        self.assertEqual([0, 1, 2], [t.pos for t in result])
        
        result = [t.copy() for t in ana(u"single")]
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].text, "single")
    
    def test_shingles(self):
        ana = RegexTokenizer(r"\w+") | ShingleFilter(3, " ")
        source = u"better a witty fool than a foolish wit"
        results = [t.copy() for t in ana(source, positions=True, chars=True)]
        self.assertEqual([t.text for t in results],
                         [u'better a witty', u'a witty fool',
                          u'witty fool than', u'fool than a', u'than a foolish',
                          u'a foolish wit'])
        self.assertEqual([t.pos for t in results], range(len(results)))
        for t in results:
            self.assertEqual(t.text, source[t.startchar:t.endchar])
        
    def test_unicode_blocks(self):
        from whoosh.support.unicode import blocks, blockname, blocknum
        
        self.assertEqual(blockname(u'a'), 'Basic Latin')
        self.assertEqual(blockname(unichr(0x0b80)), 'Tamil')
        self.assertEqual(blockname(unichr(2048)), None)
        self.assertEqual(blocknum(u'a'), 0)
        self.assertEqual(blocknum(unichr(0x0b80)), 22)
        self.assertEqual(blocknum(unichr(2048)), None)
        self.assertEqual(blocknum(u'a'), blocks.Basic_Latin)
        self.assertEqual(blocknum(unichr(0x0b80)), blocks.Tamil)
        
    def test_double_metaphone(self):
        mf = RegexTokenizer() | LowercaseFilter() | DoubleMetaphoneFilter()
        results = [(t.text, t.boost) for t in mf(u"Spruce View")]
        self.assertEqual(results, [('SPRS', 1.0), ('F', 1.0), ('FF', 0.5)])
        
        mf = RegexTokenizer() | LowercaseFilter() | DoubleMetaphoneFilter(combine=True)
        results = [(t.text, t.boost) for t in mf(u"Spruce View")]
        self.assertEqual(results, [('spruce', 1.0), ('SPRS', 1.0),
                                   ('view', 1.0), ('F', 1.0), ('FF', 0.5)])
    
        namefield = fields.TEXT(analyzer=mf)
        texts = list(namefield.process_text(u"Spruce View", mode="query"))
        self.assertEqual(texts, [u'spruce', 'SPRS', u'view', 'F', 'FF'])
    
    def test_substitution(self):
        mf = RegexTokenizer(r"\S+") | SubstitutionFilter("-", "")
        self.assertEqual([t.text for t in mf(u"one-two th-re-ee four")], ["onetwo", "threee", "four"])
        
        mf = RegexTokenizer(r"\S+") | SubstitutionFilter("([^=]*)=(.*)", r"\2=\1")
        self.assertEqual([t.text for t in mf(u"a=b c=d ef")], ["b=a", "d=c", "ef"])
    
    def test_delimited_attribute(self):
        ana = RegexTokenizer(r"\S+") | DelimitedAttributeFilter()
        results = [(t.text, t.boost) for t in ana(u"image render^2 file^0.5")]
        self.assertEqual(results, [("image", 1.0), ("render", 2.0), ("file", 0.5)])
        
    def test_porter2(self):
        from whoosh.lang.porter2 import stem
        
        plurals = ['caresses', 'flies', 'dies', 'mules', 'denied',
                   'died', 'agreed', 'owned', 'humbled', 'sized',
                   'meeting', 'stating', 'siezing', 'itemization',
                   'sensational', 'traditional', 'reference', 'colonizer',
                   'plotted']
        singles = [stem(w) for w in plurals]
        
        self.assertEqual(singles, ['caress', 'fli', 'die', 'mule', 'deni',
                                   'die', 'agre', 'own', 'humbl', 'size',
                                   'meet', 'state', 'siez', 'item', 'sensat',
                                   'tradit', 'refer', 'colon', 'plot'])
        
        self.assertEqual(stem("bill's"), "bill")
        self.assertEqual(stem("y's"), "y")
    
    def test_url(self):
        sample = u"Visit http://bitbucket.org/mchaput/whoosh or urn:isbn:5930502 or http://www.apple.com/."
        
        for ana in (SimpleAnalyzer(url_pattern),
                    StandardAnalyzer(url_pattern, stoplist=None)):
            ts = [t.text for t in ana(sample)]
            self.assertEqual(ts, [u'visit',
                                  u'http://bitbucket.org/mchaput/whoosh',
                                  u'or', u'urn:isbn:5930502', u'or',
                                  u'http://www.apple.com/'])
    
    def test_name_field(self):
        ana = (RegexTokenizer(r"\S+")
               | LowercaseFilter()
               | DoubleMetaphoneFilter(combine=True))
        namefield = fields.TEXT(analyzer=ana, multitoken_query="or")
        schema = fields.Schema(id=fields.STORED, name=namefield)
        
        ix = RamStorage().create_index(schema)
        w = ix.writer()
        w.add_document(id=u"one", name=u"Leif Ericson")
        w.commit()
        
        s = ix.searcher()
        qp = qparser.QueryParser("name", schema)
        q = qp.parse(u"leaf eriksen", normalize=False)
        r = s.search(q)
        self.assertEqual(len(r), 1)





if __name__ == '__main__':
    unittest.main()



#===============================================================================
# Copyright 2007 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""Miscellaneous utility functions and classes.
"""

from __future__ import with_statement
import codecs
import re
import sys
import time
from array import array
from copy import copy
from functools import wraps
from math import log
from struct import pack, unpack
from threading import Lock

from whoosh.system import IS_LITTLE


try:
    from itertools import permutations
except ImportError:
    # This function was only added to itertools in 2.6...
    def permutations(iterable, r=None):
        pool = tuple(iterable)
        n = len(pool)
        r = n if r is None else r
        if r > n:
            return
        indices = range(n)
        cycles = range(n, n - r, -1)
        yield tuple(pool[i] for i in indices[:r])
        while n:
            for i in reversed(range(r)):
                cycles[i] -= 1
                if cycles[i] == 0:
                    indices[i:] = indices[i + 1:] + indices[i:i + 1]
                    cycles[i] = n - i
                else:
                    j = cycles[i]
                    indices[i], indices[-j] = indices[-j], indices[i]
                    yield tuple(pool[i] for i in indices[:r])
                    break
            else:
                return


if sys.platform == 'win32':
    now = time.clock
else:
    now = time.time


# Note: these functions return a tuple of (text, length), so when you call
# them, you have to add [0] on the end, e.g. str = utf8encode(unicode)[0]

utf8encode = codecs.getencoder("utf_8")
utf8decode = codecs.getdecoder("utf_8")


# Functions

def array_to_string(a):
    if IS_LITTLE:
        a = copy(a)
        a.byteswap()
    return a.tostring()


def string_to_array(typecode, s):
    a = array(typecode)
    a.fromstring(s)
    if IS_LITTLE:
        a.byteswap()
    return a


def make_binary_tree(fn, args, **kwargs):
    """Takes a function/class that takes two positional arguments and a list of
    arguments and returns a binary tree of instances.
    
    >>> make_binary_tree(UnionMatcher, [matcher1, matcher2, matcher3])
    UnionMatcher(matcher1, UnionMatcher(matcher2, matcher3))
    
    Any keyword arguments given to this function are passed to the class
    initializer.
    """
    
    count = len(args)
    if not count:
        raise ValueError("Called make_binary_tree with empty list")
    elif count == 1:
        return args[0]
    
    half = count // 2
    return fn(make_binary_tree(fn, args[:half], **kwargs),
              make_binary_tree(fn, args[half:], **kwargs), **kwargs)


# Varint cache

# Build a cache of the varint byte sequences for the first N integers, so we
# don't have to constantly recalculate them on the fly. This makes a small but
# noticeable difference.

def _varint(i):
    s = ""
    while (i & ~0x7F) != 0:
        s += chr((i & 0x7F) | 0x80)
        i = i >> 7
    s += chr(i)
    return s


_varint_cache_size = 512
_varint_cache = []
for i in xrange(0, _varint_cache_size):
    _varint_cache.append(_varint(i))
_varint_cache = tuple(_varint_cache)


def varint(i):
    """Encodes the given integer into a string of the minimum number  of bytes.
    """
    if i < len(_varint_cache):
        return _varint_cache[i]
    return _varint(i)


def varint_to_int(vi):
    b = ord(vi[0])
    p = 1
    i = b & 0x7f
    shift = 7
    while b & 0x80 != 0:
        b = ord(vi[p])
        p += 1
        i |= (b & 0x7F) << shift
        shift += 7
    return i


def signed_varint(i):
    """Zig-zag encodes a signed integer into a varint.
    """
    
    if i >= 0:
        return varint(i << 1)
    return varint((i << 1) ^ (~0))


def decode_signed_varint(i):
    """Zig-zag decodes an integer value.
    """
    
    if not i & 1:
        return i >> 1
    return (i >> 1) ^ (~0)


def read_varint(readfn):
    """
    Reads a variable-length encoded integer.
    
    :param readfn: a callable that reads a given number of bytes,
        like file.read().
    """

    b = ord(readfn(1))
    i = b & 0x7F

    shift = 7
    while b & 0x80 != 0:
        b = ord(readfn(1))
        i |= (b & 0x7F) << shift
        shift += 7
    return i


# Fibonacci function

_fib_cache = {}


def fib(n):
    """Returns the nth value in the Fibonacci sequence.
    """

    if n <= 2:
        return n
    if n in _fib_cache:
        return _fib_cache[n]
    result = fib(n - 1) + fib(n - 2)
    _fib_cache[n] = result
    return result


# Float-to-byte encoding/decoding

def float_to_byte(value, mantissabits=5, zeroexp=2):
    """Encodes a floating point number in a single byte.
    """

    # Assume int size == float size

    fzero = (63 - zeroexp) << mantissabits
    bits = unpack("i", pack("f", value))[0]
    smallfloat = bits >> (24 - mantissabits)
    if smallfloat < fzero:
        # Map negative numbers and 0 to 0
        # Map underflow to next smallest non-zero number
        if bits <= 0:
            return chr(0)
        else:
            return chr(1)
    elif smallfloat >= fzero + 0x100:
        # Map overflow to largest number
        return chr(255)
    else:
        return chr(smallfloat - fzero)


def byte_to_float(b, mantissabits=5, zeroexp=2):
    """Decodes a floating point number stored in a single byte.
    """
    b = ord(b)
    if b == 0:
        return 0.0

    bits = (b & 0xff) << (24 - mantissabits)
    bits += (63 - zeroexp) << 24
    return unpack("f", pack("i", bits))[0]


# Length-to-byte approximation functions

def length_to_byte(length):
    """Returns a logarithmic approximation of the given number, in the range
    0-255. The approximation has high precision at the low end (e.g.
    1 -> 0, 2 -> 1, 3 -> 2 ...) and low precision at the high end. Numbers
    equal to or greater than 108116 all approximate to 255.
    
    This is useful for storing field lengths, where the general case is small
    documents and very large documents are more rare.
    """
    
    # This encoding formula works up to 108116 -> 255, so if the length is
    # equal to or greater than that limit, just return 255.
    if length >= 108116:
        return 255
    
    # The parameters of this formula where chosen heuristically so that low
    # numbers would approximate closely, and the byte range 0-255 would cover
    # a decent range of document lengths (i.e. 1 to ~100000).
    return int(round(log((length / 27.0) + 1, 1.033)))


def _byte_to_length(n):
    return int(round((pow(1.033, n) - 1) * 27))

_length_byte_cache = array("i", (_byte_to_length(i) for i in xrange(256)))
byte_to_length = _length_byte_cache.__getitem__


# Prefix encoding functions

def first_diff(a, b):
    """Returns the position of the first differing character in the strings
    a and b. For example, first_diff('render', 'rending') == 4. This function
    limits the return value to 255 so the difference can be encoded in a single
    byte.
    """

    i = -1
    for i in xrange(0, len(a)):
        if a[i] != b[1]:
            return i
        if i == 255:
            return i


def prefix_encode(a, b):
    """Compresses string b as an integer (encoded in a byte) representing
    the prefix it shares with a, followed by the suffix encoded as UTF-8.
    """
    i = first_diff(a, b)
    return chr(i) + b[i:].encode("utf8")


def prefix_encode_all(ls):
    """Compresses the given list of (unicode) strings by storing each string
    (except the first one) as an integer (encoded in a byte) representing
    the prefix it shares with its predecessor, followed by the suffix encoded
    as UTF-8.
    """

    last = u''
    for w in ls:
        i = first_diff(last, w)
        yield chr(i) + w[i:].encode("utf8")
        last = w


def prefix_decode_all(ls):
    """Decompresses a list of strings compressed by prefix_encode().
    """

    last = u''
    for w in ls:
        i = ord(w[0])
        decoded = last[:i] + w[1:].decode("utf8")
        yield decoded
        last = decoded


# Natural key sorting function

_nkre = re.compile(r"\D+|\d+", re.UNICODE)


def _nkconv(i):
    try:
        return int(i)
    except ValueError:
        return i.lower()


def natural_key(s):
    """Converts string ``s`` into a tuple that will sort "naturally" (i.e.,
    ``name5`` will come before ``name10`` and ``1`` will come before ``A``).
    This function is designed to be used as the ``key`` argument to sorting
    functions.
    
    :param s: the str/unicode string to convert.
    :rtype: tuple
    """

    # Use _nkre to split the input string into a sequence of
    # digit runs and non-digit runs. Then use _nkconv() to convert
    # the digit runs into ints and the non-digit runs to lowercase.
    return tuple(_nkconv(m) for m in _nkre.findall(s))


# Mixins and decorators

class ClosableMixin(object):
    """Mix-in for classes with a close() method to allow them to be used as a
    context manager.
    """

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()


def protected(func):
    """Decorator for storage-access methods. This decorator (a) checks if the
    object has already been closed, and (b) synchronizes on a threading lock.
    The parent object must have 'is_closed' and '_sync_lock' attributes.
    """

    @wraps(func)
    def protected_wrapper(self, *args, **kwargs):
        if self.is_closed:
            raise Exception("%r has been closed" % self)
        with self._sync_lock:
            return func(self, *args, **kwargs)

    return protected_wrapper
    

def synchronized(func):
    """Decorator for storage-access methods, which synchronizes on a threading
    lock. The parent object must have 'is_closed' and '_sync_lock' attributes.
    """
    
    @wraps(func)
    def synchronized_wrapper(self, *args, **kwargs):
        with self._sync_lock:
            return func(self, *args, **kwargs)

    return synchronized_wrapper


def unbound_cache(func):
    """Caching decorator with an unbounded cache size.
    """
    
    cache = {}
    
    @wraps(func)
    def caching_wrapper(*args):
        try:
            return cache[args]
        except KeyError:
            result = func(*args)
            cache[args] = result
            return result
        
    return caching_wrapper


def lru_cache(maxsize=100):
    """Least-recently-used cache decorator.

    This function duplicates (more-or-less) the protocol of the
    ``functools.lru_cache`` decorator in the Python 3.2 standard library, but
    uses the clock face LRU algorithm instead of an ordered dictionary.

    If *maxsize* is set to None, the LRU features are disabled and the cache
    can grow without bound.

    Arguments to the cached function must be hashable.

    View the cache statistics named tuple (hits, misses, maxsize, currsize)
    with f.cache_info().  Clear the cache and statistics with f.cache_clear().
    Access the underlying function with f.__wrapped__.
    """
    
    def decorating_function(user_function):

        stats = [0, 0, 0]  # hits, misses
        data = {}
        
        if maxsize:
            # The keys at each point on the clock face
            clock_keys = [None] * maxsize
            # The "referenced" bits at each point on the clock face
            clock_refs = array("B", (0 for _ in xrange(maxsize)))
            lock = Lock()
            
            @wraps(user_function)
            def wrapper(*args):
                key = args
                try:
                    with lock:
                        pos, result = data[key]
                        # The key is in the cache. Set the key's reference bit
                        clock_refs[pos] = 1
                        # Record a cache hit
                        stats[0] += 1
                except KeyError:
                    # Compute the value
                    result = user_function(*args)
                    with lock:
                        # Current position of the clock hand
                        hand = stats[2]
                        # Remember to stop here after a full revolution
                        end = hand
                        # Sweep around the clock looking for a position with
                        # the reference bit off
                        while True:
                            hand = (hand + 1) % maxsize
                            current_ref = clock_refs[hand]
                            if current_ref:
                                # This position's "referenced" bit is set. Turn
                                # the bit off and move on.
                                clock_refs[hand] = 0
                            elif not current_ref or hand == end:
                                # We've either found a position with the
                                # "reference" bit off or reached the end of the
                                # circular cache. So we'll replace this
                                # position with the new key
                                current_key = clock_keys[hand]
                                if current_key in data:
                                    del data[current_key]
                                clock_keys[hand] = key
                                clock_refs[hand] = 1
                                break
                        # Put the key and result in the cache
                        data[key] = (hand, result)
                        # Save the new hand position
                        stats[2] = hand
                        # Record a cache miss
                        stats[1] += 1
                return result
        
        else:
            @wraps(user_function)
            def wrapper(*args):
                key = args
                try:
                    result = data[key]
                    stats[0] += 1
                except KeyError:
                    result = user_function(*args)
                    data[key] = result
                    stats[1] += 1
                return result

        def cache_info():
            """Report cache statistics"""
            return (stats[0], stats[1], maxsize, len(data))

        def cache_clear():
            """Clear the cache and cache statistics"""
            data.clear()
            stats[0] = stats[1] = stats[2] = 0
            for i in xrange(maxsize):
                clock_keys[i] = None
                clock_refs[i] = 0

        wrapper.cache_info = cache_info
        wrapper.cache_clear = cache_clear
        return wrapper

    return decorating_function


def find_object(name, blacklist=None, whitelist=None):
    """Imports and returns an object given a fully qualified name.
    
    >>> find_object("whoosh.analysis.StopFilter")
    <class 'whoosh.analysis.StopFilter'>
    """
    
    if blacklist:
        for pre in blacklist:
            if name.startswith(pre):
                raise TypeError("%r: can't instantiate names starting with %r" % (name, pre))
    if whitelist:
        passes = False
        for pre in whitelist:
            if name.startswith(pre):
                passes = True
                break
        if not passes:
            raise TypeError("Can't instantiate %r" % name)
    
    lastdot = name.rfind(".")
    
    assert lastdot > -1, "Name %r must be fully qualified" % name
    modname = name[:lastdot]
    clsname = name[lastdot + 1:]
    
    mod = __import__(modname, fromlist=[clsname])
    cls = getattr(mod, clsname)
    return cls




















#!python

import os.path, sys
from setuptools import setup, find_packages

sys.path.insert(0, os.path.abspath("src"))
from whoosh import __version__, versionstring

setup(
	name = "Whoosh",
	version = versionstring(),
	package_dir = {'': 'src'},
	packages = ["whoosh", "whoosh.filedb", "whoosh.lang", "whoosh.qparser", "whoosh.support"],
	
	author = "Matt Chaput",
	author_email = "matt@whoosh.ca",
	
	description = "Fast, pure-Python full text indexing, search, and spell checking library.",
    long_description = open("README.txt").read(),

	license = "Apache 2.0",
	keywords = "index search text spell",
	url = "http://bitbucket.org/mchaput/whoosh",
	
	zip_safe = True,
	test_suite = "tests",
	
	classifiers = [
	"Development Status :: 5 - Production/Stable",
	"Intended Audience :: Developers",
	"License :: OSI Approved :: Apache Software License",
	"Natural Language :: English",
	"Operating System :: OS Independent",
	"Programming Language :: Python :: 2.5",
	"Topic :: Software Development :: Libraries :: Python Modules",
	"Topic :: Text Processing :: Indexing",
	],
)
